# from jax import config
# config.update("jax_enable_x64",True)

import os
import sys
import time

import chex
import flax
import flax.linen as nn
import gymnax
import jax
import jax.numpy as jnp
import numpy as np
import optax
from flax.linen.initializers import constant, orthogonal
from flax.training.train_state import TrainState
from typing import Any, Dict, NamedTuple, Sequence
import distrax
from gymnax.environments import spaces

sys.path.append('../purejaxrl')
sys.path.append('../AlphaTrade')
#Code snippet to disable all jitting.
from jax import config

from gymnax_exchange.jaxen.exec_env import ExecutionEnv

config.update("jax_disable_jit", False) 
# config.update("jax_disable_jit", True)
config.update("jax_check_tracer_leaks",False) #finds a whole assortment of leaks if true... bizarre.

import datetime
wandbOn = True
# wandbOn = False
if wandbOn:
    import wandb
    

from purejaxrl.experimental.s5.s5 import StackedEncoderModel, init_S5SSM, make_DPLR_HiPPO
from purejaxrl.experimental.s5.wrappers import FlattenObservationWrapper, LogWrapper


def save_checkpoint(params, filename):
    with open(filename, 'wb') as f:
        f.write(flax.serialization.to_bytes(params))
        print(f"Checkpoint saved to {filename}")


d_model = 256
ssm_size = 256
C_init = "lecun_normal"
discretization="zoh"
dt_min=0.001
dt_max=0.1
n_layers = 4
conj_sym=True
clip_eigs=False
bidirectional=False

blocks = 1
block_size = int(ssm_size / blocks)

Lambda, _, B, V, B_orig = make_DPLR_HiPPO(ssm_size)

block_size = block_size // 2
ssm_size = ssm_size // 2

Lambda = Lambda[:block_size]
V = V[:, :block_size]

Vinv = V.conj().T


ssm_init_fn = init_S5SSM(H=d_model,
                            P=ssm_size,
                            Lambda_re_init=Lambda.real,
                            Lambda_im_init=Lambda.imag,
                            V=V,
                            Vinv=Vinv,
                            C_init=C_init,
                            discretization=discretization,
                            dt_min=dt_min,
                            dt_max=dt_max,
                            conj_sym=conj_sym,
                            clip_eigs=clip_eigs,
                            bidirectional=bidirectional)

class ActorCriticS5(nn.Module):
    action_dim: Sequence[int]
    config: Dict

    def setup(self):
        self.encoder_0 = nn.Dense(128, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0))
        self.encoder_1 = nn.Dense(256, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0))
    
        self.action_body_0 = nn.Dense(128, kernel_init=orthogonal(2), bias_init=constant(0.0))
        self.action_body_1 = nn.Dense(128, kernel_init=orthogonal(2), bias_init=constant(0.0))
        self.action_decoder = nn.Dense(self.action_dim, kernel_init=orthogonal(0.01), bias_init=constant(0.5))

        self.value_body_0 = nn.Dense(128, kernel_init=orthogonal(2), bias_init=constant(0.0))
        self.value_body_1 = nn.Dense(128, kernel_init=orthogonal(2), bias_init=constant(0.0))
        self.value_decoder = nn.Dense(1, kernel_init=orthogonal(1.0), bias_init=constant(0.0))

        self.s5 = StackedEncoderModel(
            ssm=ssm_init_fn,
            d_model=d_model,
            n_layers=n_layers,
            activation="half_glu1",
        )
        self.actor_logtstd = self.param("log_std", nn.initializers.constant(-0.7), (self.action_dim,))


    def __call__(self, hidden, x):
        obs, dones = x
        embedding = self.encoder_0(obs)
        embedding = nn.leaky_relu(embedding)
        embedding = self.encoder_1(embedding)
        embedding = nn.leaky_relu(embedding)

        hidden, embedding = self.s5(hidden, embedding, dones)

        actor_mean = self.action_body_0(embedding)
        actor_mean = nn.leaky_relu(actor_mean)
        actor_mean = self.action_body_1(actor_mean)
        actor_mean = nn.leaky_relu(actor_mean)
        actor_mean = self.action_decoder(actor_mean)


        #pi = distrax.Categorical(logits=actor_mean)
        #Old version ^^
        # actor_logtstd = self.param("log_std", nn.initializers.zeros, (self.action_dim,))
        pi = distrax.MultivariateNormalDiag(actor_mean, jnp.exp(self.actor_logtstd))
        #New version ^^

        critic = self.value_body_0(embedding)
        critic = nn.leaky_relu(critic)
        critic = self.value_body_1(critic)
        critic = nn.leaky_relu(critic)
        critic = self.value_decoder(critic)

        return hidden, pi, jnp.squeeze(critic, axis=-1)

class Transition(NamedTuple):
    done: jnp.ndarray
    action: jnp.ndarray
    value: jnp.ndarray
    reward: jnp.ndarray
    log_prob: jnp.ndarray
    obs: jnp.ndarray
    info: jnp.ndarray


def make_train(config):
    config["NUM_UPDATES"] = (
        config["TOTAL_TIMESTEPS"] // config["NUM_STEPS"] // config["NUM_ENVS"]
    )
    config["MINIBATCH_SIZE"] = (
        config["NUM_ENVS"] * config["NUM_STEPS"] // config["NUM_MINIBATCHES"]
    )

    # # old version
    # env, env_params = gymnax.make(config["ENV_NAME"])
    # env = FlattenObservationWrapper(env)
    # env = LogWrapper(env)
    # # old version
    
    # new version
    # env= ExecutionEnv(config["ATFOLDER"],config["TASKSIDE"],config["TASK_SIZE"],config["LAMBDA"],config["GAMMA"])
    env= ExecutionEnv(config["ATFOLDER"],config["TASKSIDE"],config["WINDOW_INDEX"],config["ACTION_TYPE"],config["TASK_SIZE"],config["REWARD_LAMBDA"])
    env_params = env.default_params
    env = LogWrapper(env)
    
    # #FIXME : Uncomment normalisation.
    # if config["NORMALIZE_ENV"]:
    #     env = NormalizeVecObservation(env)
    #     env = NormalizeVecReward(env, config["GAMMA"])
    # new version

    def linear_schedule(count):
        frac = (
            1.0
            - (count // (config["NUM_MINIBATCHES"] * config["UPDATE_EPOCHS"]))
            / config["NUM_UPDATES"]
        )
        return config["LR"] * frac

    def train(rng):
        # INIT NETWORK
        
        # old version 
        # network = ActorCriticS5(env.action_space(env_params).n, config=config)
        network = ActorCriticS5(env.action_space(env_params).shape[0], config=config)
        # new version
        
        rng, _rng = jax.random.split(rng)
        init_x = (
            jnp.zeros(
                (1, config["NUM_ENVS"], *env.observation_space(env_params).shape)
            ),
            jnp.zeros((1, config["NUM_ENVS"])),
        )
        init_hstate = StackedEncoderModel.initialize_carry(config["NUM_ENVS"], ssm_size, n_layers)
        network_params = network.init(_rng, init_hstate, init_x)
        if config["ANNEAL_LR"]:
            tx = optax.chain(
                optax.clip_by_global_norm(config["MAX_GRAD_NORM"]),
                optax.adam(learning_rate=linear_schedule,b1=0.9,b2=0.99, eps=1e-5),
            )
        else:
            tx = optax.chain(
                optax.clip_by_global_norm(config["MAX_GRAD_NORM"]),
                optax.adam(config["LR"],b1=0.9,b2=0.99, eps=1e-5),
            )
        train_state = TrainState.create(
            apply_fn=network.apply,
            params=network_params,
            tx=tx,
        )

        # INIT ENV
        rng, _rng = jax.random.split(rng)
        reset_rng = jax.random.split(_rng, config["NUM_ENVS"])
        obsv, env_state = jax.vmap(env.reset, in_axes=(0, None))(reset_rng, env_params)
        init_hstate = StackedEncoderModel.initialize_carry(config["NUM_ENVS"], ssm_size, n_layers)

        # TRAIN LOOP
        def _update_step(runner_state, unused):
            # COLLECT TRAJECTORIES
            def _env_step(runner_state, unused):
                train_state, env_state, last_obs, last_done, hstate, rng = runner_state
                rng, _rng = jax.random.split(rng)

                # SELECT ACTION
                ac_in = (last_obs[np.newaxis, :], last_done[np.newaxis, :])
                hstate, pi, value = network.apply(train_state.params, hstate, ac_in)
                # sampled = pi.sample(seed=_rng)
                # action = jnp.clip(sampled, -5, 5)
                action = pi.sample(seed=_rng)
                # jax.debug.print("sampled \n {}", sampled)
                log_prob = pi.log_prob(action)
                # jax.debug.print("log_prob \n {}", log_prob)
                value, action, log_prob = (
                    value.squeeze(0),
                    action.squeeze(0),
                    log_prob.squeeze(0),
                )

                # STEP ENV
                rng, _rng = jax.random.split(rng)
                rng_step = jax.random.split(_rng, config["NUM_ENVS"])
                obsv, env_state, reward, done, info = jax.vmap(
                    env.step, in_axes=(0, 0, 0, None)
                )(rng_step, env_state, action, env_params)
                transition = Transition(
                    last_done, action, value, reward, log_prob, last_obs, info
                )
                runner_state = (train_state, env_state, obsv, done, hstate, rng)
                return runner_state, transition

            initial_hstate = runner_state[-2]
            runner_state, traj_batch = jax.lax.scan(
                _env_step, runner_state, None, config["NUM_STEPS"]
            )

            # CALCULATE ADVANTAGE
            train_state, env_state, last_obs, last_done, hstate, rng = runner_state
            ac_in = (last_obs[np.newaxis, :], last_done[np.newaxis, :])
            _, _, last_val = network.apply(train_state.params, hstate, ac_in)
            last_val = last_val.squeeze(0)
            def _calculate_gae(traj_batch, last_val, last_done):
                def _get_advantages(carry, transition):
                    gae, next_value, next_done = carry
                    done, value, reward = transition.done, transition.value, transition.reward 
                    delta = reward + config["GAMMA"] * next_value * (1 - next_done) - value
                    gae = delta + config["GAMMA"] * config["GAE_LAMBDA"] * (1 - next_done) * gae
                    return (gae, value, done), gae
                _, advantages = jax.lax.scan(_get_advantages, (jnp.zeros_like(last_val), last_val, last_done), traj_batch, reverse=True, unroll=16)
                return advantages, advantages + traj_batch.value
            advantages, targets = _calculate_gae(traj_batch, last_val, last_done)

            # UPDATE NETWORK
            def _update_epoch(update_state, unused):
                def _update_minbatch(train_state, batch_info):
                    init_hstate, traj_batch, advantages, targets = batch_info

                    def _loss_fn(params, init_hstate, traj_batch, gae, targets):
                        # RERUN NETWORK
                        _, pi, value = network.apply(
                            params, init_hstate, (traj_batch.obs, traj_batch.done)
                        )
                        log_prob = pi.log_prob(traj_batch.action)

                        # CALCULATE VALUE LOSS
                        value_pred_clipped = traj_batch.value + (
                            value - traj_batch.value
                        ).clip(-config["CLIP_EPS"], config["CLIP_EPS"])
                        value_losses = jnp.square(value - targets)
                        value_losses_clipped = jnp.square(value_pred_clipped - targets)
                        value_loss = (
                            0.5 * jnp.maximum(value_losses, value_losses_clipped).mean()
                        )

                        # CALCULATE ACTOR LOSS
                        ratio = jnp.exp(log_prob - traj_batch.log_prob)
                        gae = (gae - gae.mean()) / (gae.std() + 1e-8)
                        loss_actor1 = ratio * gae
                        loss_actor2 = (
                            jnp.clip(
                                ratio,
                                1.0 - config["CLIP_EPS"],
                                1.0 + config["CLIP_EPS"],
                            )
                            * gae
                        )
                        loss_actor = -jnp.minimum(loss_actor1, loss_actor2)
                        loss_actor = loss_actor.mean()
                        entropy = pi.entropy().mean()

                        total_loss = (
                            loss_actor
                            + config["VF_COEF"] * value_loss
                            - config["ENT_COEF"] * entropy
                        )
                        return total_loss, (value_loss, loss_actor, entropy)

                    grad_fn = jax.value_and_grad(_loss_fn, has_aux=True)
                    total_loss, grads = grad_fn(
                        train_state.params, init_hstate, traj_batch, advantages, targets
                    )
                    train_state = train_state.apply_gradients(grads=grads)
                    return train_state, total_loss

                (
                    train_state,
                    init_hstate,
                    traj_batch,
                    advantages,
                    targets,
                    rng,
                ) = update_state

                rng, _rng = jax.random.split(rng)
                permutation = jax.random.permutation(_rng, config["NUM_ENVS"])
                batch = (init_hstate, traj_batch, advantages, targets)

                shuffled_batch = jax.tree_util.tree_map(
                    lambda x: jnp.take(x, permutation, axis=1), batch
                )

                minibatches = jax.tree_util.tree_map(
                    lambda x: jnp.swapaxes(
                        jnp.reshape(
                            x,
                            [x.shape[0], config["NUM_MINIBATCHES"], -1]
                            + list(x.shape[2:]),
                        ),
                        1,
                        0,
                    ),
                    shuffled_batch,
                )

                train_state, total_loss = jax.lax.scan(
                    _update_minbatch, train_state, minibatches
                )
                update_state = (
                    train_state,
                    init_hstate,
                    traj_batch,
                    advantages,
                    targets,
                    rng,
                )
                return update_state, total_loss

            init_hstate = initial_hstate # TBH
            update_state = (
                train_state,
                init_hstate,
                traj_batch,
                advantages,
                targets,
                rng,
            )
            update_state, loss_info = jax.lax.scan(
                _update_epoch, update_state, None, config["UPDATE_EPOCHS"]
            )
            # grad_norm=jnp.mean(loss_info[1])
            train_state = update_state[0]
            # metric = (traj_batch.info,train_state.params,grad_norm)
            metric = (traj_batch.info,train_state.params)
            rng = update_state[-1]
            if config.get("DEBUG"):
                
                # # old version
                # def callback(info):
                #     return_values = info["returned_episode_returns"][info["returned_episode"]]
                #     timesteps = info["timestep"][info["returned_episode"]] * config["NUM_ENVS"]
                #     for t in range(len(timesteps)):
                #         print(f"global step={timesteps[t]}, episodic return={return_values[t]}")
                        

                # new version
                def callback(metric):
                    
                    info,trainstate_params=metric
                    # info,trainstate,grad_norm=metric
                    
                    return_values = info["returned_episode_returns"][info["returned_episode"]]
                    timesteps = info["timestep"][info["returned_episode"]] * config["NUM_ENVS"]
                    
                    def evaluation():
                        if not os.path.exists(config['CHECKPOINT_DIR']): os.makedirs(config['CHECKPOINT_DIR'])
                        # Inside your loop or function where you save the checkpoint
                        if any(timesteps % int(1e3) == 0) and len(timesteps) > 0:  # +1 since global_step is 0-indexed
                            start = time.time()
                            jax.debug.print(">>> checkpoint saving {}",round(timesteps[0], -3))
                            # Save the checkpoint to the specific directory
                            checkpoint_filename = os.path.join(config['CHECKPOINT_DIR'], f"checkpoint_{round(timesteps[0], -3)}.ckpt")
                            save_checkpoint(trainstate_params, checkpoint_filename)  # Assuming trainstate_params contains your model's state
                            jax.debug.print("+++ checkpoint saved  {}",round(timesteps[0], -3))
                            jax.debug.print("+++ time taken        {}",time.time()-start)        
                    evaluation()
                    
                    revenues = info["total_revenue"][info["returned_episode"]]
                    quant_executed = info["quant_executed"][info["returned_episode"]]
                    average_price = info["average_price"][info["returned_episode"]]
                    
                    slippage_rm = info["slippage_rm"][info["returned_episode"]]
                    price_drift_rm = info["price_drift_rm"][info["returned_episode"]]
                    price_adv_rm = info["price_adv_rm"][info["returned_episode"]]
                    vwap_rm = info["vwap_rm"][info["returned_episode"]]
                    
                    current_step = info["current_step"][info["returned_episode"]]
                    advantage_reward = info["advantage_reward"][info["returned_episode"]]
                    
                    for t in range(len(timesteps)):  
                        if wandbOn:
                            wandb.log(
                                {
                                    "global_step": timesteps[t],
                                    "episodic_return": return_values[t],
                                    "episodic_revenue": revenues[t],
                                    "quant_executed":quant_executed[t],
                                    "average_price":average_price[t],
                                    "slippage_rm":slippage_rm[t],
                                    "price_adv_rm":price_adv_rm[t],
                                    "price_drift_rm":price_drift_rm[t],
                                    "vwap_rm":vwap_rm[t],
                                    "current_step":current_step[t],
                                    "advantage_reward":advantage_reward[t],
                                    # "grad_norm":grad_norm,
                                }
                            ) 
                            print(
                                f"global step={timesteps[t]:<8} | episodic return={return_values[t]:<15} | episodic revenue={revenues[t]:<15} | average_price={average_price[t]:<20}",\
                                file=open(config['RESULTS_FILE'],'a')
                            )                                 
                        else:
                            print(
                                f"global step={timesteps[t]:<8} | episodic return={return_values[t]:<15} | episodic revenue={revenues[t]:<15} | average_price={average_price[t]:<20}"
                            )
                            # print(grad_norm)     
                # new version
                
                jax.debug.callback(callback, metric)

            runner_state = (train_state, env_state, last_obs, last_done, hstate, rng)
            return runner_state, metric

        rng, _rng = jax.random.split(rng)
        runner_state = (
            train_state,
            env_state,
            obsv,
            jnp.zeros((config["NUM_ENVS"]), dtype=bool),
            init_hstate,
            _rng,
        )
        runner_state, metric = jax.lax.scan(
            _update_step, runner_state, None, config["NUM_UPDATES"]
        )
        return {"runner_state": runner_state, "metric": metric}

    return train



    

if __name__ == "__main__":
    # try:
    #     ATFolder = sys.argv[1] 
    # except:
    #     ATFolder = "/homes/80/kang/AlphaTrade/training_oneDay"
    #     # ATFolder = '/homes/80/kang/AlphaTrade'
    #     # ATFolder = '/home/duser/AlphaTrade'
    # print("AlphaTrade folder:",ATFolder)
    timestamp=datetime.datetime.now().strftime("%m-%d_%H-%M")

    ppo_config = {
        # "LR": 2.5e-3,
        # "LR": 2.5e-4,
        "LR": 2.5e-5,
        # "LR": 2.5e-6,
        "ENT_COEF": 0.1,
        # "ENT_COEF": 0.01,
        "NUM_ENVS": 500,
        "TOTAL_TIMESTEPS": 1e8,
        # "TOTAL_TIMESTEPS": 1e7,
        # "TOTAL_TIMESTEPS": 3.5e7,
        "NUM_MINIBATCHES": 2,
        # "NUM_MINIBATCHES": 4,
        "UPDATE_EPOCHS": 5,
        # "UPDATE_EPOCHS": 4,
        "NUM_STEPS": 455,
        # "NUM_STEPS": 10,
        "CLIP_EPS": 0.2,
        # "CLIP_EPS": 0.2,
        
        "GAMMA": 0.99,
        "GAE_LAMBDA": 0.95,
        "VF_COEF": 0.5,
        "MAX_GRAD_NORM": 2.0,
        "ANNEAL_LR": True,
        "NORMALIZE_ENV": False,
        
        "ACTOR_TYPE":"S5",
        
        "ENV_NAME": "alphatradeExec-v0",
        # "WINDOW_INDEX": 0,
        "WINDOW_INDEX": -1,
        "DEBUG": True,
        "ATFOLDER": ".",
        "TASKSIDE":'sell',
        "REWARD_LAMBDA":1,
        "ACTION_TYPE":"pure",
        # "ACTION_TYPE":"delta",
        "TASK_SIZE":500,
        "RESULTS_FILE":"./results_file_"+f"{timestamp}",
        "CHECKPOINT_DIR":"./checkpoints_"+f"{timestamp}",
    }

    if wandbOn:
        run = wandb.init(
            project="AlphaTradeJAX_Train",
            config=ppo_config,
            # sync_tensorboard=True,  # auto-upload  tensorboard metrics
            save_code=True,  # optional
        )
        import datetime;params_file_name = f'params_file_{wandb.run.name}_{timestamp}'
        print(f"Results would be saved to {params_file_name}")
    else:
        import datetime;params_file_name = f'params_file_{timestamp}'
        print(f"Results would be saved to {params_file_name}")
        



    device = jax.devices()[0]
    # device = jax.devices()[1]
    # device = jax.devices()[-1]
    rng = jax.device_put(jax.random.PRNGKey(0), device)
    train_jit = jax.jit(make_train(ppo_config), device=device)
    out = train_jit(rng)

    # if jax.device_count() == 1:
    #     # +++++ Single GPU +++++
    #     rng = jax.random.PRNGKey(0)
    #     # rng = jax.random.PRNGKey(30)
    #     train_jit = jax.jit(make_train(ppo_config))
    #     start=time.time()
    #     out = train_jit(rng)
    #     print("Time: ", time.time()-start)
    #     # +++++ Single GPU +++++
    # else:
    #     # +++++ Multiple GPUs +++++
    #     num_devices = int(jax.device_count())
    #     rng = jax.random.PRNGKey(30)
    #     rngs = jax.random.split(rng, num_devices)
    #     train_fn = lambda rng: make_train(ppo_config)(rng)
    #     start=time.time()
    #     out = jax.pmap(train_fn)(rngs)
    #     print("Time: ", time.time()-start)
    #     # +++++ Multiple GPUs +++++
    
    

    # '''
    # # ---------- Save Output ----------
    import flax

    train_state = out['runner_state'][0] # runner_state.train_state
    params = train_state.params
    


    import datetime;params_file_name = f'params_file_{wandb.run.name}_{timestamp}'
    # Save the params to a file using flax.serialization.to_bytes
    with open(params_file_name, 'wb') as f:
        f.write(flax.serialization.to_bytes(params))
        print(f"pramas saved")

    # Load the params from the file using flax.serialization.from_bytes
    with open(params_file_name, 'rb') as f:
        restored_params = flax.serialization.from_bytes(flax.core.frozen_dict.FrozenDict, f.read())
        print(f"pramas restored")
        
    # jax.debug.breakpoint()
    # assert jax.tree_util.tree_all(jax.tree_map(lambda x, y: (x == y).all(), params, restored_params))
    # print(">>>")
    # '''

    if wandbOn:
        run.finish()
        
        
        



import jax
import jax.numpy as jnp
import gymnax
import sys
sys.path.append('/Users/sasrey/AlphaTrade')
sys.path.append('/homes/80/kang/AlphaTrade')
from gymnax_exchange.jaxen.base_env import BaseLOBEnv
from gymnax_exchange.jaxob import JaxOrderBookArrays as job
import chex
import time

import faulthandler

faulthandler.enable()
print("Num Jax Devices:",jax.device_count(),"Device List:",jax.devices())

chex.assert_gpu_available(backend=None)

#Code snippet to disable all jitting.
#from jax import config
#config.update("jax_disable_jit", True)


if __name__ == "__main__":
    try:
        ATFolder = sys.argv[1]
        print("AlphaTrade folder:",ATFolder)
    except:
        ATFolder = '/homes/80/kang/AlphaTrade'

    # enable_vmap=True 
    # enable_2nd_singles=True
    enable_vmap=False 
    enable_2nd_singles=False

    rng = jax.random.PRNGKey(234)
    rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)

    env=BaseLOBEnv(ATFolder) 
    env_params=env.default_params
    print(env_params.message_data.shape, env_params.book_data.shape)

    
    start=time.time()
    obs,state=env.reset(key_reset,env_params)
    print("State after reset: \n",state)
    print("Time for reset: \n",time.time()-start)
    print(env_params.message_data.shape, env_params.book_data.shape)

    if enable_2nd_singles:
        #jax.profiler.start_trace("/tmp/tensorboard")
        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        obs.block_until_ready()
        #jax.profiler.stop_trace()
        print("State after reset: \n",state)
        print("Time for 2nd reset: \n",time.time()-start)
        print(env_params.message_data.shape, env_params.book_data.shape)



    """test_action={"sides":jnp.array([1,1,1]),
                 "quantities":jnp.array([10,10,10]),
                 "prices":jnp.array([2154900,2154000,2153900]),
                 }"""
    
    test_action=env.action_space().sample(key_policy)
    print("Sampled actions are: ",test_action)

    start=time.time()
    obs,state,reward,done,info=env.step(key_step, state,test_action, env_params)
    print("State after one step: \n",state,done)
    print("Time for one step: \n",time.time()-start)

    if enable_2nd_singles:
        test_action=env.action_space().sample(key_policy)
        print("Sampled actions are: \n",test_action)
        
        start=time.time()
        obs,state,reward,done,info=env.step(key_step, state,test_action, env_params)
        print("State after 2 steps: \n",state,done)
        print("Time for 2nd step: \n",time.time()-start)
        #comment


    ####### Testing the vmap abilities ########
    
    if enable_vmap:
        vmap_reset = jax.vmap(env.reset, in_axes=(0, None))
        vmap_step = jax.jit(jax.vmap(env.step, in_axes=(0, 0, 0, None)))
        
        vmap_act_sample=jax.vmap(env.action_space().sample, in_axes=(0))

        num_envs = 100000
        vmap_keys = jax.random.split(rng, num_envs)

        test_actions=vmap_act_sample(vmap_keys)
        print(test_actions)

        start=time.time()
        obs, state = vmap_reset(vmap_keys, env_params)
        print("Time for vmap reset with,",num_envs, " environments : \n",time.time()-start)
        
        n_obs, n_state, reward, done, _ = vmap_step(vmap_keys, state, test_actions, env_params)

        start=time.time()
        for i in range(10):
            n_obs, n_state, reward, done, _ = vmap_step(vmap_keys, state, test_actions, env_params)
        print("Time for vmap step with,",num_envs, " environments : \n",time.time()-start)



import os
import sys
import time
import csv
import datetime
from os import listdir
from os.path import isfile,join

import jax
import flax
import numpy as np
import jax.numpy as jnp

sys.path.append('/homes/80/kang/AlphaTrade')
from gymnax_exchange.jaxrl.ppoS5ExecCont import ActorCriticS5
from gymnax_exchange.jaxrl.ppoS5ExecCont import StackedEncoderModel, ssm_size, n_layers
from purejaxrl.experimental.s5.s5 import StackedEncoderModel, init_S5SSM, make_DPLR_HiPPO
from gymnax_exchange.jaxen.exec_env import *

ppo_config = {    
    "ENV_NAME": "alphatradeExec-v0",
    "DEBUG": True,
    "TASKSIDE":'sell',

    "REWARD_LAMBDA":0,
    "ACTION_TYPE":"pure",
    # "ACTION_TYPE":"delta",
    "TASK_SIZE":500,
    
    "WINDOW_INDEX": -1,
    "ATFOLDER": "/homes/80/kang/AlphaTrade/testing_oneDay",
    "RESULTS_FILE":"/homes/80/kang/AlphaTrade/results_file_"+f"{datetime.datetime.now().strftime('%m-%d_%H-%M')}",
    
    
    "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-19_06-27/", # N.O. 23, pure quant
    "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-19_06-27/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-15_00-33/", # N.O. 11, pure quant
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-15_00-33/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-15_10-03/", # N.O. 11, pure quant
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-15_10-03/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-14_10-16/", # N.O. 10, pure quant
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-14_10-16/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-11_04-22/", # N.O. 3, pure quant
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-11_04-22/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/ckpt/",
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/ckpt/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-07_09-09/",
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-07_09-09/csv/",
    # "CHECKPOINT_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-06_12-57/",
    # "CHECKPOINT_CSV_DIR":"/homes/80/kang/AlphaTrade/checkpoints_10-06_12-57/csv/",
}

dir = ppo_config['CHECKPOINT_DIR']
csv_dir = ppo_config["CHECKPOINT_CSV_DIR"]


def make_evaluation(network_config):    
    network,trainstate_params,checkpoint,env,env_params,key_step = network_config
    def step(rng, obs, done, hstate,state):
        
        ac_in = (obs[np.newaxis,np.newaxis, :], jnp.array([done])[np.newaxis, :])
        assert len(ac_in[0].shape) == 3, f"{ac_in[0].shape}"
        assert len(ac_in[1].shape) == 2, f"{ac_in[1].shape}"
        hstate, pi, value = network.apply(trainstate_params, hstate, ac_in) 
        raw_action = pi.sample(seed=rng)
        # action = raw_action.round().astype(jnp.int32)[0,0,:].clip(0, None)
        action = raw_action[0,0,:]
        obs,state,reward,done,info=env.step(key_step, state,action, env_params)
        row_data = [
            checkpoint, info['window_index'], info['current_step'], info['average_price'], action.sum(), action[0], action[1],raw_action[0,0,0],raw_action[0,0,1],
            info['done'], info['slippage'], info['price_drift'], info['advantage_reward'], 
            info['drift_reward'], info['quant_executed'], 
            info['task_to_execute'], info['total_revenue']
        ]
        return row_data,(obs, done, hstate,state,reward,info)
    return step


    
    
def evaluate_savefile(paramsFile,window_idx):
    # env=ExecutionEnv(ppo_config['ATFOLDER'],ppo_config["TASKSIDE"],window_idx)
    env= ExecutionEnv(ppo_config['ATFOLDER'],ppo_config["TASKSIDE"],window_idx,ppo_config["ACTION_TYPE"],ppo_config["TASK_SIZE"],ppo_config["REWARD_LAMBDA"])
    env_params=env.default_params
    assert env.task_size == 500
    # Automatically create the directory if it doesn't exist
    os.makedirs(csv_dir, exist_ok=True)
    with open(csv_dir+paramsFile.split(".")[0]+f'_wdw_idx_{window_idx}.csv', 'w', newline='') as csvfile:
        print(paramsFile)
        csvwriter = csv.writer(csvfile)
        # Add a header row if needed
        row_title = [
            'checkpiont_name','window_index', 'current_step' , 'average_price', 'delta_sum',"delta_aggressive",'delta_passive','raw_delta_aggressive','raw_delta_passive','done', 'slippage', 'price_drift', 'advantage_reward', 'drift_reward','quant_executed', 'task_to_execute', 'total_revenue'
        ]
        csvwriter.writerow(row_title)
        csvfile.flush() 
    
        with open(dir+paramsFile, 'rb') as f:
            trainstate_params = flax.serialization.from_bytes(flax.core.frozen_dict.FrozenDict, f.read())
            print(f"pramas restored")
        rng = jax.random.PRNGKey(0)
        rng, key_reset, key_step = jax.random.split(rng, 3)
        obs,state=env.reset(key_reset,env_params)
        network = ActorCriticS5(env.action_space(env_params).shape[0], config=ppo_config)
        hstate = StackedEncoderModel.initialize_carry(1, ssm_size, n_layers)
        done = False
        
        checkpoint = int(paramsFile.split("_")[1].split(".")[0])
        network_config = (network,trainstate_params,checkpoint,env,env_params,key_step)
        device = jax.devices()[-1]
        evaluate_jit = jax.jit(make_evaluation(network_config), device=device)    
        
        rng = jax.device_put(jax.random.PRNGKey(0), device)
        for i in range(1,10000):
            print(i)

            # start = time.time()
            rng, _rng = jax.random.split(rng)
            row_data,(obs, done, hstate,state,reward,info) = evaluate_jit(rng, obs,done,hstate,state)
            # print(f"time taken: {time.time()-start}")
            
            csvwriter.writerow(row_data)
            csvfile.flush() 
            if done:
                break
            
def twap_evaluation(paramsFile,window_idx):
    env=ExecutionEnv(ppo_config['ATFOLDER'],ppo_config["TASKSIDE"],window_idx)
    env_params=env.default_params
    assert env.task_size == 500
    # Automatically create the directory if it doesn't exist
    os.makedirs(csv_dir, exist_ok=True)    
    with open(csv_dir+paramsFile.split(".")[0]+f'_twap_{window_idx}.csv', 'w', newline='') as csvfile:
        print(paramsFile)
        csvwriter = csv.writer(csvfile)
        # Add a header row if needed
        row_title = [
            'checkpiont_name','window_index', 'current_step' , 'average_price', 'delta_sum',"delta_aggressive",'delta_passive','raw_delta_aggressive','raw_delta_passive','done', 'slippage', 'price_drift', 'advantage_reward', 'drift_reward','quant_executed', 'task_to_execute', 'total_revenue'
        ]
        csvwriter.writerow(row_title)
        csvfile.flush() 
    
        rng = jax.random.PRNGKey(0)
        rng, key_reset, key_step = jax.random.split(rng, 3)
        obs,state=env.reset(key_reset,env_params)
        done = False
        for i in range(1,10000):
            print(i)
            raw_action = jnp.array([0, 0])
            # action = raw_action.round().astype(jnp.int32)[0,0,:].clip(0, None)
            action = raw_action
            obs,state,reward,done,info=env.step(key_step, state,action, env_params)
            row_data = [
                paramsFile.split("_")[1].split(".")[0], info['window_index'], info['current_step'], info['average_price'], action.sum(), action[0], action[1],raw_action[0],raw_action[1],
                info['done'], info['slippage'], info['price_drift'], info['advantage_reward'], 
                info['drift_reward'], info['quant_executed'], 
                info['task_to_execute'], info['total_revenue']
            ]
            csvwriter.writerow(row_data)
            csvfile.flush() 
            if done:
                break
    
import re
import argparse

def main(idx=-1):
    start_time =time.time()
    for window_idx in range(13):
        start=time.time()
        print(f">>> window_idx: {window_idx}")
        def extract_number_from_filename(filename):
            match = re.search(r'_(\d+)', filename)
            if match:
                return int(match.group(1))
            return 0  # default if no number is found

        onlyfiles = [f for f in listdir(dir) if isfile(join(dir, f))]
        onlyfiles = sorted(onlyfiles, key=extract_number_from_filename)
        paramsFile = onlyfiles[idx]
        evaluate_savefile(paramsFile,window_idx)
        print(f"Time for evaluation: \n",time.time()-start)
    print(f"Total time for evaluation ppo : \n",time.time()-start_time)
    
def main2(idx=-1):
    start_time =time.time()
    for window_idx in range(13):
        start=time.time()
        print(f">>> window_idx: {window_idx}")        
        def extract_number_from_filename(filename):
            match = re.search(r'_(\d+)', filename)
            if match:
                return int(match.group(1))
            return 0  # default if no number is found

        onlyfiles = [f for f in listdir(dir) if isfile(join(dir, f))]
        onlyfiles = sorted(onlyfiles, key=extract_number_from_filename)
        paramsFile = onlyfiles[idx]
        twap_evaluation(paramsFile,window_idx)
        print(f"Time for evaluation: \n",time.time()-start)
    print(f"Total time for evaluation twap: \n",time.time()-start_time)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Specify index of the file to evaluate.')
    parser.add_argument('--idx', metavar='idx', type=int, default=-1, help='Index of the file to evaluate.')
    
    args = parser.parse_args()
    main(args.idx)
    # main2(args.idx)
    # /bin/python3 /homes/80/kang/AlphaTrade/gymnax_exchange/test_scripts/evaluation.py 2
    # /bin/python3 /homes/80/kang/AlphaTrade/gymnax_exchange/test_scripts/evaluation.py -1



"""Module containing all functions to manipulate the orderbook.

To follow the functional programming paradigm of JAX, the functions of the
orderbook are not put into an object but are left standalone.

The functions exported are:

add_order : Adds a given order to the given orderside 
cancel_order : Removes quantity (and order if remainder <0) from a given order side
match_order : remove quantity from a single standing order and generate trades.



"""


from textwrap import fill
from typing import Optional, OrderedDict
from jax import numpy as jnp
import jax
from functools import partial, partialmethod
import chex

INITID = -9000
#MAXPRICE=999999999
#TODO: Get rid of these magic numbers by allowing a config dict to be passed through as a static arg 

MAX_INT = 2_147_483_647  # max 32 bit int

############### ADD AND REMOVE ###############
@jax.jit
def add_order(orderside: chex.Array, msg: dict) -> chex.Array :
    """Low level function that will add an order (Dict)
      to the orderbook (Array) and return the updated"""
    emptyidx=jnp.where(orderside==-1,size=1,fill_value=-1)[0]
    orderside=orderside.at[emptyidx,:].set(jnp.array([msg['price'],jnp.maximum(0,msg['quantity']),msg['orderid'],msg['traderid'],msg['time'],msg['time_ns']])).astype(jnp.int32)
    return __removeZeroNegQuant(orderside)

@jax.jit
def __removeZeroNegQuant(orderside):
    return jnp.where((orderside[:,1]<=0).reshape((orderside.shape[0],1)),x=(jnp.ones(orderside.shape)*-1).astype(jnp.int32),y=orderside)


# @jax.jit
# def cancel_order(orderside,msg):
#     # jax.debug.breakpoint()
#     condition=((orderside[:,2]==msg['orderid']) | ((orderside[:,0]==msg['price']) & (orderside[:,2]<=-9000)))
#     idx=jnp.where(condition,size=1,fill_value=-1)[0]
#     orderside=orderside.at[idx,1].set(orderside[idx,1]-msg['quantity'])
#     return __removeZeroNegQuant(orderside)

@jax.jit
def cancel_order(orderside, msg):
    def get_init_id_match(orderside, msg):
        init_id_match = ((orderside[:, 0] == msg['price']) & (orderside[:, 2] <= INITID))
        idx = jnp.where(init_id_match, size=1, fill_value=-1)[0][0]
        return idx

    # jax.debug.breakpoint()
    # TODO: also check for price here?
    oid_match = (orderside[:, 2] == msg['orderid'])
    idx = jnp.where(oid_match, size=1, fill_value=-1)[0][0]
    idx = jax.lax.cond(idx == -1, get_init_id_match, lambda a, b: idx, orderside, msg)
    orderside = orderside.at[idx, 1].set(orderside[idx, 1] - msg['quantity'])
    return __removeZeroNegQuant(orderside)

############### MATCHING FUNCTIONS ###############

@jax.jit
def match_bid_order(data_tuple):
    return __get_top_bid_order_idx(data_tuple[1]), *match_order(data_tuple)

@jax.jit
def match_ask_order(data_tuple):
    return __get_top_ask_order_idx(data_tuple[1]), *match_order(data_tuple)

@jax.jit
def match_order(data_tuple):
    top_order_idx, orderside, qtm, price, trade, agrOID, time, time_ns = data_tuple
    newquant=jnp.maximum(0,orderside[top_order_idx,1]-qtm) #Could theoretically be removed as an operation because the removeZeroQuand func also removes negatives. 
    qtm=qtm-orderside[top_order_idx,1]
    qtm=qtm.astype(jnp.int32)
    emptyidx=jnp.where(trade==-1,size=1,fill_value=-1)[0]
    trade=trade.at[emptyidx,:].set(jnp.array([orderside[top_order_idx,0],orderside[top_order_idx,1]-newquant,orderside[top_order_idx,2],[agrOID],[time],[time_ns]]).transpose())
    orderside=__removeZeroNegQuant(orderside.at[top_order_idx,1].set(newquant))
    return (orderside.astype(jnp.int32), jnp.squeeze(qtm), price, trade, agrOID, time, time_ns)

@jax.jit
def __get_top_bid_order_idx(orderside):
    maxPrice=jnp.max(orderside[:,0],axis=0)
    times=jnp.where(orderside[:,0]==maxPrice,orderside[:,4],999999999)
    minTime_s=jnp.min(times,axis=0)
    times_ns=jnp.where(times==minTime_s,orderside[:,5],999999999)
    minTime_ns=jnp.min(times_ns,axis=0)
    return jnp.where(times_ns==minTime_ns,size=1,fill_value=-1)[0]

@jax.jit
def __get_top_ask_order_idx(orderside):
    prices=orderside[:,0]
    prices=jnp.where(prices==-1,999999999,prices)
    minPrice=jnp.min(prices)
    times=jnp.where(orderside[:,0]==minPrice,orderside[:,4],999999999)
    minTime_s=jnp.min(times,axis=0)
    times_ns=jnp.where(times==minTime_s,orderside[:,5],999999999)
    minTime_ns=jnp.min(times_ns,axis=0)
    return jnp.where(times_ns==minTime_ns,size=1,fill_value=-1)[0]

@jax.jit
def __check_before_matching_bid(data_tuple):
    top_order_idx,orderside,qtm,price,trade,_,_,_=data_tuple
    returnarray=(orderside[top_order_idx,0]>=price) & (qtm>0) & (orderside[top_order_idx,0]!=-1)
    return jnp.squeeze(returnarray)

@jax.jit
def _match_against_bid_orders(orderside,qtm,price,trade,agrOID,time,time_ns):
    top_order_idx=__get_top_bid_order_idx(orderside)
    top_order_idx,orderside,qtm,price,trade,_,_,_=jax.lax.while_loop(__check_before_matching_bid,match_bid_order,(top_order_idx,orderside,qtm,price,trade,agrOID,time,time_ns))
    return (orderside,qtm,price,trade)

@jax.jit
def __check_before_matching_ask(data_tuple):
    top_order_idx,orderside,qtm,price,trade,_,_,_=data_tuple
    returnarray=(orderside[top_order_idx,0]<=price) & (qtm>0) & (orderside[top_order_idx,0]!=-1)
    return jnp.squeeze(returnarray)

@jax.jit
def _match_against_ask_orders(orderside,qtm,price,trade,agrOID,time,time_ns):
    top_order_idx=__get_top_ask_order_idx(orderside)
    top_order_idx,orderside,qtm,price,trade,_,_,_=jax.lax.while_loop(__check_before_matching_ask,match_ask_order,(top_order_idx,orderside,qtm,price,trade,agrOID,time,time_ns))
    return (orderside,qtm,price,trade)

########Type Functions#############

def doNothing(msg,askside,bidside,trades):
    return askside,bidside,trades

def bid_lim(msg,askside,bidside,trades):
    #match with asks side
    #add remainder to bids side
    matchtuple=_match_against_ask_orders(askside,msg["quantity"],msg["price"],trades,msg['orderid'],msg["time"],msg["time_ns"])
    #^(orderside,qtm,price,trade)
    msg["quantity"]=matchtuple[1]
    bids=add_order(bidside,msg)
    return matchtuple[0],bids,matchtuple[3]

def bid_cancel(msg,askside,bidside,trades):
    return askside,cancel_order(bidside,msg),trades

def bid_mkt(msg,askside,bidside,trades):
    msg["price"]=999999999
    matchtuple=_match_against_ask_orders(askside,msg["quantity"],msg["price"],trades,msg['orderid'],msg["time"],msg["time_ns"])
    #^(orderside,qtm,price,trade)
    return matchtuple[0],bidside,matchtuple[3]


def ask_lim(msg,askside,bidside,trades):
    #match with bids side
    #add remainder to asks side
    matchtuple=_match_against_bid_orders(bidside,msg["quantity"],msg["price"],trades,msg['orderid'],msg["time"],msg["time_ns"])
    #^(orderside,qtm,price,trade)
    msg["quantity"]=matchtuple[1]
    asks=add_order(askside,msg)
    return asks,matchtuple[0],matchtuple[3]

def ask_cancel(msg,askside,bidside,trades):
    return cancel_order(askside,msg),bidside,trades

def ask_mkt(msg,askside,bidside,trades):
    #set price to 0
    #match with bids side 
    #no need to add remainder
    msg["price"]=0
    matchtuple=_match_against_bid_orders(bidside,msg["quantity"],msg["price"],trades,msg['orderid'],msg["time"],msg["time_ns"])
    #^(orderside,qtm,price,trade)
    return askside,matchtuple[0],matchtuple[3]


############### MAIN BRANCHING FUNCS ###############

@jax.jit
def cond_type_side(book_state, data):
    askside,bidside,trades=book_state
    #jax.debug.breakpoint()
    #jax.debug.print("Askside before is \n {}",askside)
    msg={
    'side':data[1],
    'type':data[0],
    'price':data[3],
    'quantity':data[2],
    'orderid':data[5],
    'traderid':data[4],
    'time':data[6],
    'time_ns':data[7]}
    # index=((msg["side"]+1)+msg["type"]).astype(jnp.int32)
    s = msg["side"]
    t = msg["type"]
    index = (((s == -1) & (t == 1)) | ((s ==  1) & (t == 4))) * 0 + \
            (((s ==  1) & (t == 1)) | ((s == -1) & (t == 4))) * 1 + \
            (((s == -1) & (t == 2)) | ((s == -1) & (t == 3))) * 2 + \
            (((s ==  1) & (t == 2)) | ((s ==  1) & (t == 3))) * 3
    # jax.debug.print("msg[side] {}", msg["side"])
    # jax.debug.print("msg[type] {}", msg["type"])
    # jax.debug.print("index is {}", index)
    # ask,bid,trade=jax.lax.switch(index-1,(ask_lim,ask_cancel,bid_lim,bid_cancel),msg,askside,bidside,trades)
    ask, bid, trade = jax.lax.switch(index, (ask_lim, bid_lim, ask_cancel, bid_cancel), msg, askside, bidside, trades)
    return (ask, bid, trade), 0

@jax.jit
def cond_type_side_save_states(ordersides,data):
    askside,bidside,trades=ordersides
    #jax.debug.breakpoint()
    #jax.debug.print("Askside before is \n {}",askside)
    msg={
    'side':data[1],
    'type':data[0],
    'price':data[3],
    'quantity':data[2],
    'orderid':data[5],
    'traderid':data[4],
    'time':data[6],
    'time_ns':data[7]}
    # index=((msg["side"]+1)+msg["type"]).astype(jnp.int32)
    s = msg["side"]
    t = msg["type"]
    index = (((s == -1) & (t == 1)) | ((s ==  1) & (t == 4))) * 0 + \
            (((s ==  1) & (t == 1)) | ((s == -1) & (t == 4))) * 1 + \
            (((s == -1) & (t == 2)) | ((s == -1) & (t == 3))) * 2 + \
            (((s ==  1) & (t == 2)) | ((s ==  1) & (t == 3))) * 3
    # ask,bid,trade=jax.lax.switch(index-1,(ask_lim,ask_cancel,bid_lim,bid_cancel),msg,askside,bidside,trades)
    ask, bid, trade = jax.lax.switch(index, (ask_lim, bid_lim, ask_cancel, bid_cancel), msg, askside, bidside, trades)
    #jax.debug.print("Askside after is \n {}",ask)
    return (ask,bid,trade),(ask,bid,trade)

@jax.jit
def cond_type_side_save_bidask(ordersides,data):
    askside,bidside,trades=ordersides
    #jax.debug.breakpoint()
    #jax.debug.print("Askside before is \n {}",askside)
    msg={
    'side':data[1],
    'type':data[0],
    'price':data[3],
    'quantity':data[2],
    'orderid':data[5],
    'traderid':data[4],
    'time':data[6],
    'time_ns':data[7]}
    # index=((msg["side"]+1)+msg["type"]).astype(jnp.int32)
    s = msg["side"]
    t = msg["type"]
    index = (((s == -1) & (t == 1)) | ((s ==  1) & (t == 4))) * 0 + \
            (((s ==  1) & (t == 1)) | ((s == -1) & (t == 4))) * 1 + \
            (((s == -1) & (t == 2)) | ((s == -1) & (t == 3))) * 2 + \
            (((s ==  1) & (t == 2)) | ((s ==  1) & (t == 3))) * 3
    # ask,bid,trade=jax.lax.switch(index-1,(ask_lim,ask_cancel,bid_lim,bid_cancel),msg,askside,bidside,trades)
    ask, bid, trade = jax.lax.switch(index, (ask_lim, bid_lim, ask_cancel, bid_cancel), msg, askside, bidside, trades)
    #jax.debug.print("Askside after is \n {}",ask)
    # jax.debug.breakpoint()
    return (ask,bid,trade),get_best_bid_and_ask_inclQuants(ask,bid)

vcond_type_side=jax.vmap(cond_type_side,((0,0,0),0))

############### SCAN FUNCTIONS ###############

def scan_through_entire_array(msg_array,ordersides):
    ordersides,_=jax.lax.scan(cond_type_side,ordersides,msg_array)
    return ordersides

def scan_through_entire_array_save_states(msg_array,ordersides,steplines):
    #Will return the states for each of the processed messages, but only those from data to keep array size constant, and enabling variable #of actions (AutoCancel)
    last,all=jax.lax.scan(cond_type_side_save_states,ordersides,msg_array)
    return (all[0][-steplines:],all[1][-steplines:],last[2])

def scan_through_entire_array_save_bidask(msg_array,ordersides,steplines):
    #Will return the states for each of the processed messages, but only those from data to keep array size constant, and enabling variable #of actions (AutoCancel)
    last,all=jax.lax.scan(cond_type_side_save_bidask,ordersides,msg_array)
    # jax.debug.breakpoint()
    return (last[0],last[1],last[2],all[0][-steplines:],all[1][-steplines:])

vscan_through_entire_array=jax.vmap(scan_through_entire_array,(2,(0,0,0)),0)

################ GET CANCEL MESSAGES ################

#Obtain messages to cancel based on a given ID to lookup. Currently only used in the execution environment.
def get_size(bookside,agentID):
    return jnp.sum(jnp.where(bookside[:,3]==agentID,1,0)).astype(jnp.int32)

def getCancelMsgs(bookside,agentID,size,side):
    bookside=jnp.concatenate([bookside,jnp.zeros((1,6),dtype=jnp.int32)],axis=0)
    indeces_to_cancel=jnp.where(bookside[:,3]==agentID,size=size,fill_value=-1)
    cancel_msgs=jnp.concatenate([jnp.ones((1,size),dtype=jnp.int32)*2,\
                                 jnp.ones((1,size),dtype=jnp.int32)*side, \
                                bookside[indeces_to_cancel,1], \
                                bookside[indeces_to_cancel,0], \
                                bookside[indeces_to_cancel,3], \
                                bookside[indeces_to_cancel,2], \
                                bookside[indeces_to_cancel,4], \
                                bookside[indeces_to_cancel,5]],axis=0).transpose()
    return cancel_msgs


#TODO Currently not implemented: should be used for a less naive version of the autocancel
def getCancelMsgs_smart(bookside,agentID,size,side,action_msgs):
    cond=jnp.stack([bookside[:,3]==agentID]*6,axis=1)
    indeces_to_cancel=jnp.where(bookside[:,3]==agentID,size=size,fill_value=0)
    cancel_msgs=jnp.concatenate([jnp.ones((1,size),dtype=jnp.int32)*2, \
                                jnp.ones((1,size),dtype=jnp.int32)*side, \
                                bookside[indeces_to_cancel,1], \
                                bookside[indeces_to_cancel,0], \
                                bookside[indeces_to_cancel,3], \
                                bookside[indeces_to_cancel,2], \
                                bookside[indeces_to_cancel,4], \
                                bookside[indeces_to_cancel,5]],axis=0).transpose()
    cancel_msgs=jnp.where(cancel_msgs==-1,0,cancel_msgs)
    jax.lax.scan(remove_cnl_if_renewed,cancel_msgs,action_msgs)
    return cancel_msgs

#TODO Currently not implemented: should be used for a less naive version of the autocancel
def remove_cnl_if_renewed(cancel_msgs,action_msg):
    jnp.where(cancel_msgs[:,3]==action_msg[3],)
    return cancel_msgs

   





######Helper functions for getting information #######


def get_totquant_at_price(orderside,price):
        return jnp.sum(jnp.where(orderside[:,0]==price,orderside[:,1],0))

get_totquant_at_prices=jax.vmap(get_totquant_at_price,(None,0),0)

# @partial(jax.jit,static_argnums=0)
# def get_L2_state(N,asks,bids):
#     bid_prices=-jnp.unique(-bids[:,0],size=N,fill_value=1)
#     ask_prices=jnp.unique(jnp.where(asks[:,0]==-1,999999999,asks[:,0]),size=N,fill_value=999999999)
#     ask_prices=jnp.where(ask_prices==999999999,-1,ask_prices)

#     bid_quants=get_totquant_at_prices(bids,bid_prices)
#     ask_quants=get_totquant_at_prices(asks,ask_prices)
#     bid_quants=jnp.where(bid_quants<0,0,bid_quants)
#     ask_quants=jnp.where(ask_quants<0,0,ask_quants)
#     return jnp.stack((ask_prices,ask_quants,bid_prices,bid_quants),axis=1,dtype=jnp.int32)

def get_best_bid_and_ask(asks,bids):
    best_ask=jnp.min(jnp.where(asks[:,0]==-1,999999999,asks[:,0]))
    best_bid=jnp.max(bids[:,0])
    # jax.debug.print("-----")
    # jax.debug.print("best_bid from [get_best_bid_and_ask] {}", best_bid)
    # jax.debug.print("bids {}", bids)
    return best_ask,best_bid

def get_best_bid_and_ask_inclQuants(asks,bids):
    best_ask,best_bid=get_best_bid_and_ask(asks,bids)
    best_ask_Q=jnp.sum(jnp.where(asks[:,0]==best_ask,asks[:,1],0))                     
    best_bid_Q=jnp.sum(jnp.where(bids[:,0]==best_bid,bids[:,1],0))
    best_ask=jnp.array([best_ask,best_ask_Q],dtype=jnp.int32)             
    best_bid=jnp.array([best_bid,best_bid_Q],dtype=jnp.int32)             
    return best_ask,best_bid  


@partial(jax.jit,static_argnums=0)
def init_orderside(nOrders=100):
    return (jnp.ones((nOrders,6))*-1).astype(jnp.int32)

@jax.jit
def init_msgs_from_l2(
    book_l2: jnp.ndarray,
    time: Optional[jax.Array] = None,
) -> jax.Array:
    """  """
    orderbookLevels = book_l2.shape[0] // 4  # price/quantity for bid/ask
    data = book_l2.reshape(orderbookLevels * 2, 2)
    newarr = jnp.zeros((orderbookLevels * 2, 8), dtype=jnp.int32)
    if time is None:
        time = jnp.array([34200, 0])
    initOB = newarr \
        .at[:, 3].set(data[:,0]) \
        .at[:, 2].set(data[:,1]) \
        .at[:, 0].set(1) \
        .at[0:orderbookLevels*4:2, 1].set(-1) \
        .at[1:orderbookLevels*4:2, 1].set(1) \
        .at[:, 4].set(INITID) \
        .at[:, 5].set(INITID - jnp.arange(0, orderbookLevels*2)) \
        .at[:, 6].set(time[0]) \
        .at[:, 7].set(time[1])
    return initOB


# #TODO: Actually complete this function to not only return dummy vars
# def get_initial_orders(bookData,idx_window,time):
#     orderbookLevels=10
#     initid=-9000
#     data=jnp.array(bookData[idx_window]).reshape(int(10*2),2)
#     newarr = jnp.zeros((int(orderbookLevels*2),8),dtype=jnp.int32)
#     initOB = newarr \
#         .at[:,3].set(data[:,0]) \
#         .at[:,2].set(data[:,1]) \
#         .at[:,0].set(1) \
#         .at[0:orderbookLevels*4:2,1].set(-1) \
#         .at[1:orderbookLevels*4:2,1].set(1) \
#         .at[:,4].set(initid) \
#         .at[:,5].set(initid-jnp.arange(0,orderbookLevels*2)) \
#         .at[:,6].set(time[0]) \
#         .at[:,7].set(time[1])
#     return initOB

def get_initial_time(messageData,idx_window):
    return messageData[idx_window,0,0,-2:]


def get_data_messages(messageData,idx_window,step_counter):
    messages=messageData[idx_window,step_counter,:,:]
    return messages
    




# ===================================== #
# ******* Config your own func ******** #
# ===================================== #


@jax.jit
def get_best_ask(asks):
    '''  Return the best / lowest ask price. If there is no ask, return -1. '''
    min = jnp.min(jnp.where(asks[:, 0] == -1, MAX_INT, asks[:, 0]))
    return jnp.where(min == MAX_INT, -1, min)

@jax.jit
def get_best_bid(bids):
    ''' Return the best / highest bid price. If there is no bid, return -1. '''
    return jnp.max(bids[:, 0])

@jax.jit
def get_volume_at_price(side_array, price):
    volume = jnp.sum(jnp.where(side_array[:,0] == price, side_array[:,1], 0))
    return volume

@jax.jit
def get_init_volume_at_price(
        side_array: jax.Array,
        price: int
    ) -> jax.Array:
    ''' Returns the size of initial volume (order with INITID) at given price. '''
    volume = jnp.sum(
        jnp.where(
            (side_array[:, 0] == price) & (side_array[:, 2] <= INITID), 
            side_array[:, 1], 
            0
        )
    )
    return volume

@jax.jit
def get_order_by_id(
        side_array: jax.Array,
        order_id: int,
    ) -> jax.Array:
    """ Returns all order fields for the first order matching the given order_id.
        CAVE: if the same ID is used multiple times, will only return the first 
        (e.g. for INITID)
    """
    idx = jnp.where( 
        side_array[..., 2] == order_id,
        size=1,
        fill_value=-1,
    )
    # return vector of -1 if not found
    return jax.lax.cond(
        idx == -1,
        lambda i: -1 * jnp.ones((6,), dtype=jnp.int32),
        lambda i: side_array[i][0],
        idx
    )

@jax.jit
def get_order_by_id_and_price(
        side_array: jax.Array,
        order_id: int,
        price: int,
    ) -> jax.Array:
    """ Returns all order fields for the first order matching the given order_id at the given price.
        CAVE: if the same ID is used multiple times at the given price level, will only return the first 
    """
    idx = jnp.where(
        ((side_array[..., 2] == order_id) &
         (side_array[..., 0] == price)),
        size=1,
        fill_value=-1,
    )
    # return vector of -1 if not found
    return jax.lax.cond(
        idx == -1,
        lambda i: -1 * jnp.ones((6,), dtype=jnp.int32),
        lambda i: side_array[i][0],
        idx
    )

@jax.jit
def get_order_ids(
        orderbook_array: jax.Array,
    ) -> jax.Array:
    """ Returns a list of all order ids in the orderbook
    """
    return jnp.unique(orderbook_array[:, 2], size=orderbook_array.shape[0], fill_value=1)

@partial(jax.jit, static_argnums=0)
def get_next_executable_order(side, side_array):
    # best sell order / ask
    if side == 0:
        idx = __get_top_ask_order_idx(side_array)
    # best buy order / bid
    elif side == 1:
        idx = __get_top_bid_order_idx(side_array)
    else:
        raise ValueError("Side must be 0 (bid) or 1 (ask).")
    return side_array[idx].squeeze()

@partial(jax.jit, static_argnums=2)
def get_L2_state(asks, bids, n_levels):
    # unique sorts ascending --> negative values to get descending
    bid_prices = -1 * jnp.unique(-1 * bids[:, 0], size=n_levels, fill_value=1)
    # replace -1 with max 32 bit int in sorting asks before sorting
    ask_prices = jnp.unique(
        jnp.where(asks[:, 0] == -1, MAX_INT, asks[:, 0]),
        size=n_levels,
        fill_value=-1
    )
    # replace max 32 bit int with -1 after sorting
    ask_prices = jnp.where(ask_prices == MAX_INT, -1, ask_prices)

    bids = jnp.stack((bid_prices, get_totquant_at_prices(bids, bid_prices)))
    asks = jnp.stack((ask_prices, get_totquant_at_prices(asks, ask_prices)))
    # set negative volumes to 0
    bids = bids.at[1].set(jnp.where(bids[1] < 0, 0, bids[1]))
    asks = asks.at[1].set(jnp.where(asks[1] < 0, 0, asks[1]))
    # combine asks and bids in joint representation
    l2_state = jnp.hstack((asks.T, bids.T)).flatten()
    return l2_state

vmap_get_L2_state = jax.vmap(get_L2_state, (0, 0, None), 0)



from functools import partial
import importlib
from os import remove
from readline import remove_history_item
from typing import Dict, NamedTuple, Optional, Tuple
from unicodedata import bidirectional
import gymnax_exchange.jaxob.JaxOrderBookArrays as job
job = importlib.reload(job)
import jax
from jax import numpy as jnp
from jax import lax



class LobState(NamedTuple):
    asks: jnp.ndarray
    bids: jnp.ndarray
    trades: jnp.ndarray


class OrderBook():
    def __init__(
            self: 'OrderBook',
            nOrders: int = 100,
            nTrades: int = 100
        ) -> None:
        self.nOrders = nOrders
        self.nTrades = nTrades

    @jax.jit
    def init(self: 'OrderBook') -> LobState:
        asks = (jnp.ones((self.nOrders, 6)) * -1).astype(jnp.int32)
        bids = (jnp.ones((self.nOrders, 6)) * -1).astype(jnp.int32)
        trades = (jnp.ones((self.nTrades, 6)) * -1).astype(jnp.int32)
        return LobState(asks, bids, trades)

    @jax.jit
    def reset(
            self: 'OrderBook',
            l2_book: Optional[jnp.ndarray] = None,
        ) -> LobState:
        """"""
        state = self.init()
        if l2_book is not None:
            msgs = job.init_msgs_from_l2(l2_book)
            state = self.process_orders_array(state, msgs)
        return state

    @jax.jit
    def process_order(
            self: 'OrderBook',
            state: LobState,
            quote: Dict,
            from_data: bool = False,
            verbose: bool = False
        ) -> LobState:
        '''Wrapper function for the object class that takes a Dict Object as the quote,
         ensures the order is conserved and turns the values into a jnp array which is passed to the JNP ProcessOrder function'''
        #Type, Side,quant,price
        inttype = 5
        intside = -1
        if quote['side'] == 'bid':
            intside = 1 

        if quote['type'] == 'limit':
            inttype = 1
        elif quote['type'] == 'cancel':
            inttype = 2
        elif quote['type'] == 'delete':
            inttype = 2
        elif quote['type'] == 'market':
            inttype = 1
            intside = intside * -1

        msg = jnp.array([
            inttype,
            intside,
            quote['quantity'],
            quote['price'],
            quote['trade_id'],
            quote['order_id'],
            int(quote['timestamp'].split('.')[0]),
            int(quote['timestamp'].split('.')[1])
        ], dtype=jnp.int32)

        ordersides, _ = job.cond_type_side(state, msg)
        return LobState(*ordersides)

    @jax.jit
    def process_order_array(
            self: 'OrderBook',
            state: LobState,
            quote: jax.Array,
            from_data: bool = False,
            verbose: bool = False
        ) -> LobState:
        '''Same as process_order but quote is an array.'''
        ordersides, _ = job.cond_type_side(state, quote)
        return LobState(*ordersides)

    @jax.jit
    def process_orders_array(
        self: 'OrderBook',
        state: LobState,
        msgs: jax.Array,
    ) -> LobState:
        '''Wrapper function for the object class that takes a JNP Array of messages (Shape=Nx8), and applies them, in sequence, to the orderbook'''
        return LobState(*job.scan_through_entire_array(msgs, tuple(state)))

    @partial(jax.jit, static_argnums=(3,))
    def process_orders_array_l2(
            self: 'OrderBook',
            state: LobState,
            msgs: jax.Array,
            n_levels: int
        ) -> Tuple[LobState, jax.Array]:
        all_asks, all_bids, trades = job.scan_through_entire_array_save_states(msgs, tuple(state), msgs.shape[0])
        state = LobState(all_asks[-1], all_bids[-1], trades)
        # calculate l2 states
        l2_states = job.vmap_get_L2_state(all_asks, all_bids, n_levels)
        return state, l2_states

    @partial(jax.jit, static_argnums=(2,4))
    def get_volume_at_price(
            self: 'OrderBook',
            state: LobState,
            side: int,
            price: int,
            init_only: bool = False
        ) -> int:

        if side == 0:
            side_array = state.asks
        elif side == 1:
            side_array = state.bids
        else:
            raise ValueError('Side must be 0 or 1')

        if init_only:
            return job.get_init_volume_at_price(side_array, price)
        else:
            return job.get_volume_at_price(side_array, price)

    @partial(jax.jit, static_argnums=(2,))
    def get_best_price(
            self: 'OrderBook',
            state: LobState,
            side: int
        ) -> int:
        # sell / asks
        if side == 0:
            return self.get_best_ask(state)
        # buy / bids
        elif side == 1:
            return self.get_best_bid(state)
        else:
            raise ValueError('Side must be 0 or 1')
    
    @jax.jit
    def get_best_bid(
            self: 'OrderBook',
            state: LobState
        ) -> int:
        return job.get_best_bid(state.bids)

    @jax.jit
    def get_best_ask(
            self: 'OrderBook',
            state: LobState
        ) -> int:
        return job.get_best_ask(state.asks)

    @jax.jit
    def get_best_bid_and_ask_inclQuants(
            self: 'OrderBook',
            state: LobState
        ) -> Tuple[jax.Array, jax.Array]:
        return job.get_best_bid_and_ask_inclQuants(state.asks, state.bids)

    @partial(jax.jit, static_argnums=(2,))
    def get_L2_state(
            self: 'OrderBook',
            state: LobState,
            n_levels: int
        ) -> jax.Array:
        return job.get_L2_state(state.asks, state.bids, n_levels)
    
    @partial(jax.jit, static_argnums=(2,))
    def get_side_ids(
            self: 'OrderBook',
            state: LobState,
            side: int
        ) -> jax.Array:
        if side == 0:
            return job.get_order_ids(state.asks)
        elif side == 1:
            return job.get_order_ids(state.bids)
        else:
            raise ValueError('Side must be 0 or 1')

    @partial(jax.jit, static_argnums=(2,))
    def get_order(
            self: 'OrderBook',
            state: LobState,
            side: int,
            order_id: int,
            price: Optional[int] = None,
        ) -> jax.Array:
        ''' '''
        side_array = state.asks if side == 0 else state.bids
        if price is not None:
            return job.get_order_by_id_and_price(side_array, order_id, price)
        else:
            return job.get_order_by_id(side_array, order_id)
        
    @partial(jax.jit, static_argnums=(2,))
    def get_next_executable_order(
            self: 'OrderBook',
            state: LobState,
            side: int
        ):
        if side == 0:
            side_array = state.asks
        elif side == 1:
            side_array = state.bids
        else:
            raise ValueError('Side must be 0 or 1')
        return job.get_next_executable_order(side, side_array)
    
    #Flatten and Unflatten functions so that methods can be appropriately jitted. 
    def _tree_flatten(self: 'OrderBook'):
        children = ()  # arrays / dynamic values
        aux_data = {'nOrders': self.nOrders, 'nTrades':self.nTrades}  # static values
        return (children, aux_data)

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)


jax.tree_util.register_pytree_node(
    OrderBook,
    OrderBook._tree_flatten,
    OrderBook._tree_unflatten
)



# from jax import config
# config.update("jax_enable_x64",True)

from ast import Dict
from contextlib import nullcontext
from email import message
from random import sample
from re import L
import jax
import jax.numpy as jnp
import numpy as np
from jax import lax
from gymnax.environments import environment, spaces
from typing import Tuple, Optional
import chex
from flax import struct
from gymnax_exchange.jaxob import JaxOrderBookArrays as job



@struct.dataclass
class EnvState:
    ask_raw_orders: chex.Array
    bid_raw_orders: chex.Array
    trades: chex.Array
    init_time: chex.Array
    time: chex.Array
    customIDcounter: int
    window_index:int
    step_counter: int
    max_steps_in_episode: int

@struct.dataclass
class EnvParams:
    message_data: chex.Array
    book_data: chex.Array
    episode_time: int =  60*30 #60seconds times 30 minutes = 1800seconds
    # max_steps_in_episode: int = 100
    time_per_step: int= 0##Going forward, assume that 0 implies not to use time step?
    time_delay_obs_act: chex.Array = jnp.array([0, 0]) #0ns time delay.
    




class BaseLOBEnv(environment.Environment):
    def __init__(self,alphatradePath):
        super().__init__()
        self.sliceTimeWindow = 1800 # counted by seconds, 1800s=0.5h
        self.stepLines = 100
        self.messagePath = alphatradePath+"/data/Flow_10/"
        self.orderbookPath = alphatradePath+"/data/Book_10/"
        self.start_time = 34200  # 09:30
        self.end_time = 57600  # 16:00
        self.nOrdersPerSide=100
        self.nTradesLogged=100
        self.book_depth=10
        self.n_actions=3
        self.customIDCounter=0
        self.trader_unique_id=-9000+1
        self.tick_size=100



        # Load the data from LOBSTER
        def load_LOBSTER(sliceTimeWindow, stepLines, messagePath, orderbookPath, start_time, end_time):
            def preProcessingData_csv2pkl():
                return 0
            def load_files():
                from os import listdir; from os.path import isfile, join; import pandas as pd
                readFromPath = lambda data_path: sorted([f for f in listdir(data_path) if isfile(join(data_path, f))])
                messageFiles, orderbookFiles = readFromPath(messagePath), readFromPath(orderbookPath)
                dtype = {0: float,1: int, 2: int, 3: int, 4: int, 5: int}
                messageCSVs = [pd.read_csv(messagePath + file, usecols=range(6), dtype=dtype, header=None) for file in messageFiles if file[-3:] == "csv"]
                orderbookCSVs = [pd.read_csv(orderbookPath + file, header=None) for file in orderbookFiles if file[-3:] == "csv"]
                return messageCSVs, orderbookCSVs
            messages, orderbooks = load_files()

            def preProcessingMassegeOB(message, orderbook):
                def splitTimeStamp(m):
                    m[6] = m[0].apply(lambda x: int(x))
                    m[7] = ((m[0] - m[6]) * int(1e9)).astype(int)
                    m.columns = ['time','type','order_id','qty','price','direction','time_s','time_ns']
                    return m
                message = splitTimeStamp(message)
                def filterValid(message):
                    message = message[message.type.isin([1,2,3,4])]
                    valid_index = message.index.to_numpy()
                    message.reset_index(inplace=True,drop=True)
                    return message, valid_index
                message, valid_index = filterValid(message)
                def adjustExecutions(message):
                    message.loc[message['type'] == 4, 'direction'] *= -1
                    message.loc[message['type'] == 4, 'type'] = 1
                    return message
                message = adjustExecutions(message)
                def removeDeletes(message):
                    message.loc[message['type'] == 3, 'type'] = 2
                    return message
                message = removeDeletes(message)
                def addTraderId(message):
                    import warnings
                    from pandas.errors import SettingWithCopyWarning
                    warnings.filterwarnings('ignore', category=SettingWithCopyWarning)
                    message['trader_id'] = message['order_id']
                    return message

                message = addTraderId(message)
                orderbook.iloc[valid_index,:].reset_index(inplace=True, drop=True)
                return message,orderbook
            pairs = [preProcessingMassegeOB(message, orderbook) for message,orderbook in zip(messages,orderbooks)]
            messages, orderbooks = zip(*pairs)

            def index_of_sliceWithoutOverlap(start_time, end_time, interval):
                indices = list(range(start_time, end_time, interval))
                return indices
            indices = index_of_sliceWithoutOverlap(start_time, end_time, sliceTimeWindow)
            def sliceWithoutOverlap(message, orderbook):
                # print("start")
                def splitMessage(message, orderbook):
                    import numpy as np
                    sliced_parts = []
                    init_OBs = []
                    for i in range(len(indices) - 1):
                        start_index = indices[i]
                        end_index = indices[i + 1]
                        index_s, index_e = message[(message['time'] >= start_index) & (message['time'] < end_index)].index[[0, -1]].tolist()
                        index_e = (index_e // stepLines + 10) * stepLines + index_s % stepLines
                        assert (index_e - index_s) % stepLines == 0, 'wrong code 31'
                        sliced_part = message.loc[np.arange(index_s, index_e)]
                        sliced_parts.append(sliced_part)
                        init_OBs.append(orderbook.iloc[index_s,:])

                    # Last sliced part from last index to end_time
                    start_index = indices[i]
                    end_index = indices[i + 1]
                    index_s, index_e = message[(message['time'] >= start_index) & (message['time'] < end_index)].index[[0, -1]].tolist()
                    index_s = (index_s // stepLines - 10) * stepLines + index_e % stepLines
                    assert (index_e - index_s) % stepLines == 0, 'wrong code 32'
                    last_sliced_part = message.loc[np.arange(index_s, index_e)]
                    sliced_parts.append(last_sliced_part)
                    init_OBs.append(orderbook.iloc[index_s, :])
                    for part in sliced_parts:
                        # print("start")
                        assert part.time_s.iloc[-1] - part.time_s.iloc[0] >= sliceTimeWindow, f'wrong code 33, {part.time_s.iloc[-1] - part.time_s.iloc[0]}, {sliceTimeWindow}'
                        assert part.shape[0] % stepLines == 0, 'wrong code 34'
                    return sliced_parts, init_OBs
                sliced_parts, init_OBs = splitMessage(message, orderbook)
                def sliced2cude(sliced):
                    columns = ['type','direction','qty','price','trader_id','order_id','time_s','time_ns']
                    cube = sliced[columns].to_numpy()
                    cube = cube.reshape((-1, stepLines, 8))
                    return cube
                # def initialOrderbook():
                slicedCubes = [sliced2cude(sliced) for sliced in sliced_parts]
                # Cube: dynamic_horizon * stepLines * 8
                slicedCubes_withOB = zip(slicedCubes, init_OBs)
                return slicedCubes_withOB
            slicedCubes_withOB_list = [sliceWithoutOverlap(message, orderbook) for message,orderbook in zip(messages,orderbooks)]
            # i = 6 ; message,orderbook = messages[i],orderbooks[i]
            # slicedCubes_list(nested list), outer_layer : day, inter_later : time of the day


            def nestlist2flattenlist(nested_list):
                import itertools
                flattened_list = list(itertools.chain.from_iterable(nested_list))
                return flattened_list
            Cubes_withOB = nestlist2flattenlist(slicedCubes_withOB_list)


            max_steps_in_episode_arr = jnp.array([m.shape[0] for m,o in Cubes_withOB],jnp.int32)


            def Cubes_withOB_padding(Cubes_withOB):
                max_m = max(m.shape[0] for m, o in Cubes_withOB)
                new_Cubes_withOB = []
                for cube, OB in Cubes_withOB:
                    def padding(cube, target_shape):
                        pad_width = np.zeros((100, 8))
                        # Calculate the amount of padding required
                        padding = [(0, target_shape - cube.shape[0]), (0, 0), (0, 0)]
                        padded_cube = np.pad(cube, padding, mode='constant', constant_values=0)
                        return padded_cube
                    cube = padding(cube, max_m)
                    new_Cubes_withOB.append((cube, OB))
                return new_Cubes_withOB
            Cubes_withOB = Cubes_withOB_padding(Cubes_withOB)
            
            # # breakpoint()
            # for j in range(13):
            #     data = Cubes_withOB[j][0][:,:,3]
            #     dir = f"/homes/80/kang/AlphaTrade/testing_oneDay/prices/results_file_numpy_{j}.npy"
            #     np.save(dir, data)
            #     # breakpoint()
            # # [print(Cubes_withOB[j][0][i,:,3],file=open(f"/homes/80/kang/AlphaTrade/testing_oneDay/prices/results_file_numpy_{j}",'a')) for i in range(Cubes_withOB[0][0].shape[0]) for j in range(13)]
            
            return Cubes_withOB, max_steps_in_episode_arr

        Cubes_withOB, max_steps_in_episode_arr = load_LOBSTER(self.sliceTimeWindow,self.stepLines,self.messagePath,self.orderbookPath,self.start_time,self.end_time)

        self.max_steps_in_episode_arr = max_steps_in_episode_arr
        # # ------------------------------- TESTING ------------------------------
        # # alphatradePath = '/homes/80/kang/AlphaTrade/training_oneDay'
        # alphatradePath = '/homes/80/kang/AlphaTrade/testing_oneDay'
        # messagePath = alphatradePath+"/data/Flow_10/"
        # orderbookPath = alphatradePath+"/data/Book_10/"
        # sliceTimeWindow, stepLines, messagePath, orderbookPath, start_time, end_time=1800,100,messagePath,orderbookPath,34200,57600
        # Cubes_withOB, max_steps_in_episode_arr = load_LOBSTER(1800,100,messagePath,orderbookPath,34200,57600)
        # msgs=[jnp.array(cube) for cube, book in Cubes_withOB]
        # bks=[jnp.array(book) for cube, book in Cubes_withOB]
        # message_data, book_data = msgs[0],bks[0]
        # nOrdersPerSide, nTradesLogged, tick_size,stepLines,task_size, n_ticks_in_book= 100, 100, 100,100, 20,200
        # # ------------------------------- TESTING ------------------------------

        # # # ------------------------------- STATS ------------------------------  
        # # msg = msgs[0]
        # tvs_lst = []
        # for index,msg in enumerate(msgs):
        #     pass
        #     print(f">>> {index}")
        #     tradingVolume = lambda step:step[step[:,0]==4][:,2].sum()
        #     tvs = [tradingVolume(step) for step in msg]
        #     tvs_lst.append(tvs)
        # tvs_arr = jnp.asarray(tvs_lst)
        # tvs_arr.shape
        # tvsa_sum = tvs_arr.sum(axis=1)
        # step_mean = [row[:max_steps_in_episode_arr[index]].mean().round().astype(jnp.int32) for index,row in enumerate(tvs_arr)]
        # step_mean = jnp.asarray(step_mean)
        
        
        
        # ps_lst = []
        # for index,msg in enumerate(msgs):
        #     # pass
        #     print(f">>> {index}")
        #     # step = msg[0]
        #     # step[step[:,0]==4]
        #     def vwapPirce(step):
        #         executed = step[step[:,0]==4]
        #         return (((executed[:,2]*(executed[:,3]//100)).sum() / executed[:,2].sum())*100).astype(jnp.int32)
        #     ps = [vwapPirce(step) for step in msg]
        #     ps_lst.append(ps)
        # ps_arr = jnp.asarray(ps_lst)
        # ps_arr.shape
        # def subsample_evenly(input_list, num_samples=16):
        #     step_size = len(input_list) / num_samples
        #     subsamples = []

        #     for i in range(num_samples):
        #         index = int(round(i * step_size))
        #         subsamples.append(input_list[index])

        #     return subsamples
        # def forward_fill_jax(arr):
        #     last_valid = None
        #     for i in range(arr.shape[0]):
        #         if arr[i] != 0:
        #             last_valid = arr[i]
        #         else:
        #             if last_valid is not None:
        #                 arr = arr.at[i].set(last_valid)
        #     return arr
        # def back_fill_jax(arr):
        #     farr = jnp.flip(arr)
        #     result = forward_fill_jax(farr)
        #     fresult = jnp.flip(result)
        #     return fresult
        # def mean_fill_jax(newRow):
        #     ffilled_row = forward_fill_jax(newRow)
        #     bfilled_row = back_fill_jax(newRow)
        #     mfilled_row = jnp.round((ffilled_row+bfilled_row)/2).astype(jnp.int32)
        #     return mfilled_row
        
        
        # prices_list = []
        # for index,row in enumerate(ps_arr):
        #     newRow = row[:max_steps_in_episode_arr[index]]
        #     mfilled_row = mean_fill_jax(newRow)
        #     result =  jnp.asarray(subsample_evenly(mfilled_row, num_samples=16))
        #     prices_list.append(result)
        # prices_arr = jnp.asarray(prices_list)
        # prices_arr.shape
        # p = prices_arr[prices_arr[:,0]>=30000000]
        # p_ = p[:,:-1]
        # p_head = p[:,1:]
        # pReturnInBP = (p_head-p_)/p_*10000
        
        # m=pReturnInBP.mean(axis = 0)
        # # m=p.mean(axis = 0).round().astype(jnp.int32)
        # # s=p.std(axis = 0).round().astype(jnp.int32)
        # s=pReturnInBP.std(axis = 0)
        # m
        # s
        
        # m_Minus_s = m-s
        # m
        # m_Plus_s = m +s
        # m_Plus_s = 2*m - m_Minus_s 
        
        # tvsa_sum = tvs_arr.sum(axis=1)
        # tvsa_sum.shape
        # step_mean = [row[:max_steps_in_episode_arr[index]].mean().round().astype(jnp.int32) for index,row in enumerate(tvs_arr)]
        # step_mean = jnp.asarray(step_mean)
        # # assert step_mean.shape[0] == tvs_arr.shape[0]
        # # # ------------------------------- STATS ------------------------------            
                    
        

            
            
        #List of message cubes 
        msgs=[jnp.array(cube) for cube, book in Cubes_withOB]
        bks=[jnp.array(book) for cube, book in Cubes_withOB]

        self.messages=jnp.array(msgs)   #4D Array: (n_windows x n_steps (max) x n_messages x n_features)
        self.books=jnp.array(bks)       #2D Array: (n_windows x [4*n_depth])
      
        self.n_windows=len(self.books)
        # jax.debug.breakpoint()
        
        # ==================================================================
        # ================= CAUTION NOT BELONG TO BASE ENV =================
        # ================= EPECIALLY SUPPORT FOR EXEC ENV =================
        print("START:  pre-reset in the initialization")


        nOrdersPerSide, nTradesLogged, tick_size,stepLines,task_size,n_ticks_in_book = self.nOrdersPerSide, self.nTradesLogged, self.tick_size,self.stepLines,200, 20
        
        def get_state(message_data, book_data,max_steps_in_episode):
            time=jnp.array(message_data[0,0,-2:])
            #Get initial orders (2xNdepth)x6 based on the initial L2 orderbook for this window 
            def get_initial_orders(book_data,time):
                orderbookLevels=10
                initid=-9000
                data=jnp.array(book_data).reshape(int(10*2),2)
                newarr = jnp.zeros((int(orderbookLevels*2),8),dtype=jnp.int32)
                initOB = newarr \
                    .at[:,3].set(data[:,0]) \
                    .at[:,2].set(data[:,1]) \
                    .at[:,0].set(1) \
                    .at[0:orderbookLevels*4:2,1].set(-1) \
                    .at[1:orderbookLevels*4:2,1].set(1) \
                    .at[:,4].set(initid) \
                    .at[:,5].set(initid-jnp.arange(0,orderbookLevels*2)) \
                    .at[:,6].set(time[0]) \
                    .at[:,7].set(time[1])
                return initOB
            init_orders=get_initial_orders(book_data,time)
            #Initialise both sides of the book as being empty
            asks_raw=job.init_orderside(nOrdersPerSide)
            bids_raw=job.init_orderside(nOrdersPerSide)
            trades_init=(jnp.ones((nTradesLogged,6))*-1).astype(jnp.int32)
            #Process the initial messages through the orderbook
            ordersides=job.scan_through_entire_array(init_orders,(asks_raw,bids_raw,trades_init))

            # Mid Price after init added to env state as the initial price --> Do not at to self as this applies to all environments.
            best_ask, best_bid = job.get_best_bid_and_ask_inclQuants(ordersides[0],ordersides[1])
            M = (best_bid[0] + best_ask[0])//2//tick_size*tick_size 
            
            state = (ordersides[0],ordersides[1],ordersides[2],jnp.resize(best_ask,(stepLines,2)),jnp.resize(best_bid,(stepLines,2)),\
                time,time,0,-1,M,task_size,0,0,0,0,max_steps_in_episode)
            return state
        
        def get_obs(state):
            """Return observation from raw state trafo."""
            # ========= self.get_obs(state,params) =============
            # -----------------------1--------------------------
            best_asks=state[3][:,0]
            best_bids =state[4][:,0]
            mid_prices=(best_asks+best_bids)//2//tick_size*tick_size 
            second_passives_sell_task = best_asks+tick_size*n_ticks_in_book 
            second_passives_buy_task =  best_bids-tick_size*n_ticks_in_book
            spreads = best_asks - best_bids

            # -----------------------2--------------------------
            timeOfDay = state[6] #state.time
            deltaT = timeOfDay -  state[5] # state.init_time
            # -----------------------3--------------------------
            initPrice = state[10] # state.init_price
            priceDrift = mid_prices[-1] - initPrice
            # -----------------------4--------------------------
            # -----------------------5--------------------------
            taskSize = state[11] # state.task_to_execute
            executed_quant=state[12] # state.quant_executed
            # -----------------------7--------------------------
            bestAsksQtys = state[3][:,1]
            bestBidsQtys = state[4][:,1]
            shallowImbalance = bestAsksQtys - bestBidsQtys # ShallowImbalance
            # -----------------------8--------------------------
            step_counter=0;max_steps_in_episode=state[-1]
            # ========= self.get_obs(state,params) =============
            obs_sell = jnp.concatenate((best_bids,best_asks,mid_prices,second_passives_sell_task,spreads,timeOfDay,deltaT,jnp.array([initPrice]),jnp.array([priceDrift]),jnp.array([taskSize]),jnp.array([executed_quant]),shallowImbalance,jnp.array([step_counter]),jnp.array([max_steps_in_episode])))
            obs_buy  = jnp.concatenate((best_bids,best_asks,mid_prices,second_passives_buy_task, spreads,timeOfDay,deltaT,jnp.array([initPrice]),jnp.array([priceDrift]),jnp.array([taskSize]),jnp.array([executed_quant]),shallowImbalance,jnp.array([step_counter]),jnp.array([max_steps_in_episode])))
            # jax.debug.breakpoint()
            def obsNorm(obs):
                return jnp.concatenate((
                    obs[:400]/3.5e7, # best_bids,best_asks,mid_prices,second_passives  TODO CHANGE THIS
                    obs[400:500]/100000, # spreads
                    obs[500:501]/100000, # timeOfDay
                    obs[501:502]/1000000000, # timeOfDay
                    obs[502:503]/10,# deltaT
                    obs[503:504]/1000000000,# deltaT
                    obs[504:505]/3.5e7,# initPrice  TODO CHANGE THIS
                    obs[505:506]/100000,# priceDrift
                    obs[506:507]/500, # taskSize TODO CHANGE THIS
                    obs[507:508]/500, # executed_quant TODO CHANGE THIS
                    obs[508:608]/100, # shallowImbalance 
                    obs[608:609]/300, # step_counter TODO CHANGE THIS
                    obs[609:610]/300, # max_steps_in_episode TODO CHANGE THIS
                ))
            return obsNorm(obs_sell), obsNorm(obs_buy)
        
        def get_state_obs(message_data, book_data,max_steps_in_episode):
            state = get_state(message_data, book_data,max_steps_in_episode)
            obs_sell, obs_buy = get_obs(state)
            return state, obs_sell, obs_buy

        state_obs = [get_state_obs(Cubes_withOB[i][0], Cubes_withOB[i][1], max_steps_in_episode_arr[i]) for i in range(len(max_steps_in_episode_arr))]
        
        def state2stateArray(state):
            state_5 = jnp.hstack((state[5],state[6],state[9],state[15]))
            padded_state = jnp.pad(state_5, (0, 100 - state_5.shape[0]), constant_values=-1)[:,jnp.newaxis]
            stateArray = jnp.hstack((state[0],state[1],state[2],state[3],state[4],padded_state))
            # jax.debug.breakpoint()
            return stateArray
        
        self.stateArray_list = jnp.array([state2stateArray(state) for state, obs_sell, obs_buy in state_obs])
        self.obs_sell_list = jnp.array([jnp.array(obs_sell) for state, obs_sell, obs_buy in state_obs])
        self.obs_buy_list =  jnp.array([jnp.array(obs_buy) for state, obs_sell, obs_buy in state_obs])
        print("FINISH: pre-reset in the initialization")
        # ================= CAUTION NOT BELONG TO BASE ENV =================
        # ================= EPECIALLY SUPPORT FOR EXEC ENV =================
        # ==================================================================
        #TODO Most of the state space should be exactly the same for the base and exec env, 
        # can we think about keeping the base part seperate from the exec part? 


        print(f"Num of data_window: {self.n_windows}")



    @property
    def default_params(self) -> EnvParams:
        # Default environment parameters
        return EnvParams(self.messages,self.books)
        # return EnvParams(self.messages,self.books,self.state_list,self.obs_sell_list,self.obs_buy_list)
    # @property
    # def default_params(self) -> EnvParams:
    #     # Default environment parameters
    #     # return EnvParams(self.messages,self.books)
    #     return EnvParams(self.messages,self.books,self.state_list,self.obs_sell_list,self.obs_buy_list)


    def step_env(
        self, key: chex.PRNGKey, state: EnvState, action: Dict, params: EnvParams
    ) -> Tuple[chex.Array, EnvState, float, bool, dict]:
        #Obtain the messages for the step from the message data
        data_messages=job.get_data_messages(params.message_data,state.window_index,state.step_counter)
        #jax.debug.print("Data Messages to process \n: {}",data_messages)

        #Assumes that all actions are limit orders for the moment - get all 8 fields for each action message
        types=jnp.ones((self.n_actions,),jnp.int32)
        sides=((action["sides"]+1)/2).astype(jnp.int32)      #from action space
        prices=action["prices"]     #from action space
        quants=action["quantities"] #from action space
        trader_ids=jnp.ones((self.n_actions,),jnp.int32)*self.trader_unique_id #This agent will always have the same (unique) trader ID
        order_ids=jnp.ones((self.n_actions,),jnp.int32)*(self.trader_unique_id+state.customIDcounter)+jnp.arange(0,self.n_actions) #Each message has a unique ID
        times=jnp.resize(state.time+params.time_delay_obs_act,(self.n_actions,2)) #time from last (data) message of prev. step + some delay
        #Stack (Concatenate) the info into an array 
        action_msgs=jnp.stack([types,sides,quants,prices,trader_ids,order_ids],axis=1)
        action_msgs=jnp.concatenate([action_msgs,times],axis=1)

        #Add to the top of the data messages 
        total_messages=jnp.concatenate([action_msgs,data_messages],axis=0)
        #jax.debug.print("Step messages to process are: \n {}", total_messages)

        #Save time of final message to add to state
        time=total_messages[-1:][0][-2:]

        #Process messages of step (action+data) through the orderbook
        ordersides=job.scan_through_entire_array(total_messages,(state.ask_raw_orders,state.bid_raw_orders,state.trades))

        #Update state (ask,bid,trades,init_time,current_time,OrderID counter,window index for ep, step counter)
        state = EnvState(ordersides[0],ordersides[1],ordersides[2],state.init_time,time,state.customIDcounter+self.n_actions,\
            state.window_index,state.step_counter+1,state.max_steps_in_episode)
        done = self.is_terminal(state,params)
        reward=0
        #jax.debug.print("Final state after step: \n {}", state)
        return self.get_obs(state,params),state,reward,done,{"info":0}



    def reset_env(
        self, key: chex.PRNGKey, params: EnvParams
    ) -> Tuple[chex.Array, EnvState]:
        """Reset environment state by sampling initial position in OB."""
        idx_data_window = jax.random.randint(key, minval=0, maxval=self.n_windows, shape=())

        #Get the init time based on the first message to be processed in the first step. 
        time=job.get_initial_time(params.message_data,idx_data_window) 
        #Get initial orders (2xNdepth)x6 based on the initial L2 orderbook for this window 
        init_orders=job.get_initial_orders(params.book_data,idx_data_window,time)
        #Initialise both sides of the book as being empty
        asks_raw=job.init_orderside(self.nOrdersPerSide)
        bids_raw=job.init_orderside(self.nOrdersPerSide)
        trades_init=(jnp.ones((self.nTradesLogged,6))*-1).astype(jnp.int32)
        #Process the initial messages through the orderbook
        ordersides=job.scan_through_entire_array(init_orders,(asks_raw,bids_raw,trades_init))

        #Craft the first state
        state = EnvState(*ordersides,time,time,0,idx_data_window,0,self.max_steps_in_episode_arr[idx_data_window])

        return self.get_obs(state,params),state

    def is_terminal(self, state: EnvState, params: EnvParams) -> bool:
        """Check whether state is terminal."""
        return (state.time-state.init_time)[0]>params.episode_time

    def get_obs(self, state: EnvState, params:EnvParams) -> chex.Array:
        """Return observation from raw state trafo."""
        return job.get_L2_state(self.book_depth,state.ask_raw_orders,state.bid_raw_orders)

    @property
    def name(self) -> str:
        """Environment name."""
        return "alphatradeBase-v0"

    @property
    def num_actions(self) -> int:
        """Number of actions possible in environment."""
        return self.n_actions


    def action_space(
        self, params: Optional[EnvParams] = None
    ) -> spaces.Discrete:
        """Action space of the environment."""
        return spaces.Dict(
            {
                "sides":spaces.Box(0,2,(self.n_actions,),dtype=jnp.int32),
                "quantities":spaces.Box(0,100,(self.n_actions,),dtype=jnp.int32),
                "prices":spaces.Box(0,99999999,(self.n_actions,),dtype=jnp.int32)
            }
        )

    #TODO: define obs space (4xnDepth) array of quants&prices. Not that important right now. 
    def observation_space(self, params: EnvParams):
        """Observation space of the environment."""
        return NotImplementedError

    #FIXME:Currently this will sample absolute gibberish. Might need to subdivide the 6 (resp 5) 
    #           fields in the bid/ask arrays to return something of value. Not sure if actually needed.   
    def state_space(self, params: EnvParams) -> spaces.Dict:
        """State space of the environment."""
        return spaces.Dict(
            {
                "bids": spaces.Box(-1,999999999,shape=(6,self.nOrdersPerSide),dtype=jnp.int32),
                "asks": spaces.Box(-1,999999999,shape=(6,self.nOrdersPerSide),dtype=jnp.int32),
                "trades": spaces.Box(-1,999999999,shape=(6,self.nTradesLogged),dtype=jnp.int32),
                "time": spaces.Discrete(params.max_steps_in_episode),
            }
        )



# from jax import config
# config.update("jax_enable_x64",True)
# ============== testing scripts ===============
import jax
import jax.numpy as jnp
import sys
from re import L
import time 
sys.path.append('/Users/sasrey/AlphaTrade')
sys.path.append('/homes/80/kang/AlphaTrade')
import chex
import faulthandler; faulthandler.enable()
chex.assert_gpu_available(backend=None)
from jax import config # Code snippet to disable all jitting.
config.update("jax_disable_jit", False)
# config.update("jax_disable_jit", True)
from gymnax_exchange.jaxen.exec_env import *
import json
# ============== testing scripts ===============

paramsFile = '/homes/80/kang/AlphaTrade/params_file_smooth-paper-19_09-23_04-30'
# paramsFile = '/homes/80/kang/AlphaTrade/params_file_dutiful-thunder-5_07-21_18-48'

# def twapV3(state, env_params):
#     # ---------- ifMarketOrder ----------
#     remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
#     marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
#     ifMarketOrder = (remainingTime <= marketOrderTime)
#     # print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
#     # ---------- ifMarketOrder ----------
#     # ---------- quants ----------
#     remainedQuant = state.task_to_execute - state.quant_executed
#     remainedStep = state.max_steps_in_episode - state.step_counter
#     stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
#     limit_quants = jax.random.permutation(key_policy, jnp.array([stepQuant//2,stepQuant-stepQuant//2,stepQuant//2,stepQuant-stepQuant//2]), independent=True)
#     market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
#     quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
#     # ---------- quants ----------
#     return jnp.array(quants) 

if __name__ == "__main__":
    try:
        ATFolder = sys.argv[1]
        print("AlphaTrade folder:",ATFolder)
    except:
        # ATFolder = '/home/duser/AlphaTrade'
        # ATFolder = '/homes/80/kang/AlphaTrade'
        # ATFolder = '/homes/80/kang/AlphaTrade/testing'
        # ATFolder = '/homes/80/kang/AlphaTrade/testing_small'
        ATFolder = '/homes/80/kang/AlphaTrade/testing_oneDay'
        
    env=ExecutionEnv(ATFolder,"sell")
    env_params=env.default_params
    print(env_params.message_data.shape, env_params.book_data.shape)
    assert env.task_size == 500
    import time; timestamp = str(int(time.time()))
    
    def get_ppo_average_price(rngInitNum):
        rng = jax.random.PRNGKey(rngInitNum)
        rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
        obs,state=env.reset(key_reset,env_params)
        ppo_config = {
                "LR": 2.5e-4,
                "NUM_ENVS": 1, # CAUTION !!!
                "NUM_STEPS": 1,
                "TOTAL_TIMESTEPS": 1e8,
                "UPDATE_EPOCHS": 1,
                "NUM_MINIBATCHES": 1,
                "GAMMA": 0.99,
                "GAE_LAMBDA": 0.95,
                "CLIP_EPS": 0.2,
                "ENT_COEF": 0.01,
                "VF_COEF": 0.5,
                "MAX_GRAD_NORM": 0.5,
                "ANNEAL_LR": True,
                "DEBUG": True,
                
                "ENV_NAME": "alphatradeExec-v0",
                "NORMALIZE_ENV": False,
                "ENV_LENGTH": "oneWindow",
                "ATFOLDER": ATFolder,
                "TASKSIDE":'sell',
                "LAMBDA":0.0,
                "TASK_SIZE":500,
            }
        import flax
        from gymnax_exchange.jaxrl.ppoS5ExecCont import ActorCriticS5
        from gymnax_exchange.jaxrl.ppoS5ExecCont import StackedEncoderModel, ssm_size, n_layers
        network = ActorCriticS5(env.action_space(env_params).shape[0], config=ppo_config)
        init_hstate = StackedEncoderModel.initialize_carry(ppo_config["NUM_ENVS"], ssm_size, n_layers)
            
        # ===================================================
        # CHOICE ONE
        with open(paramsFile, 'rb') as f:
            restored_params = flax.serialization.from_bytes(flax.core.frozen_dict.FrozenDict, f.read())
            print(f"pramas restored")
        # ---------------------------------------------------
        # init_x = (
        #     jnp.zeros(
        #         (1, ppo_config["NUM_ENVS"], *env.observation_space(env_params).shape)
        #     ),
        #     jnp.zeros((1, ppo_config["NUM_ENVS"])),
        # )
        # network_params = network.init(key_policy, init_hstate, init_x)
        # restored_params = network_params
        # CHOICE OTWO
        # ===================================================
        
        
        init_done = jnp.array([False]*ppo_config["NUM_ENVS"])
        ac_in = (obs[np.newaxis, np.newaxis, :], init_done[np.newaxis, :])
        assert len(ac_in[0].shape) == 3
        hstate, pi, value = network.apply(restored_params, init_hstate, ac_in)
        print("Network Carry Initialized")
        action = pi.sample(seed=rng).round().astype(jnp.int32)[0,0,:].clip( 0, None) # CAUTION about the [0,0,:], only works for num_env=1
        print(f"-------------\nPPO 0th actions are: {action} with sum {action.sum()}")
        obs,state,reward,done,info=env.step(key_step, state, action, env_params)
        print("{" + ", ".join([f"'{k}': {v}" for k, v in info.items()]) + "}")
        excuted_list = []
        for i in range(1,10000):
            # ==================== ACTION ====================
            # ---------- acion from trained network ----------
            ac_in = (obs[np.newaxis,np.newaxis, :], jnp.array([done])[np.newaxis, :])
            assert len(ac_in[0].shape) == 3, f"{ac_in[0].shape}"
            assert len(ac_in[1].shape) == 2, f"{ac_in[1].shape}"
            hstate, pi, value = network.apply(restored_params, hstate, ac_in) 
            action = pi.sample(seed=rng).round().astype(jnp.int32)[0,0,:].clip( 0, None)
            # ---------- acion from trained network ----------
            # ==================== ACTION ====================    
            print(f"-------------\nPPO {i}th actions are: {action} with sum {action.sum()}")
            start=time.time()
            obs,state,reward,done,info=env.step(key_step, state,action, env_params)
            print(f"Time for {i} step: \n",time.time()-start)
            print("{" + ", ".join([f"'{k}': {v}" for k, v in info.items()]) + "}")
            excuted_list.append(info["quant_executed"])
            if done:
                break
        return info['window_index'],info['average_price'], excuted_list
    
    def get_twap_average_price(rngInitNum):
        rng = jax.random.PRNGKey(rngInitNum)
        rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        print("Time for reset: \n",time.time()-start)
        print(env_params.message_data.shape, env_params.book_data.shape)
        excuted_list = []
        for i in range(1,10000):
            # ==================== ACTION ====================
            # ---------- acion from given strategy  ----------
            print("---"*20)
            print("window_index ",state.window_index)
            key_policy, _ = jax.random.split(key_policy,2)
            def twap(state, env_params):
                # ---------- ifMarketOrder ----------
                remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
                marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
                ifMarketOrder = (remainingTime <= marketOrderTime)
                print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
                # ---------- ifMarketOrder ----------
                # ---------- quants ----------
                remainedQuant = state.task_to_execute - state.quant_executed
                remainedStep = state.max_steps_in_episode - state.step_counter
                stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
                limit_quants_agressive = jnp.append(jax.random.permutation(key_policy, jnp.array([stepQuant - 3*stepQuant//4,stepQuant//4, stepQuant//4]), independent=True),max(stepQuant - 3*stepQuant//4,stepQuant//4))
                limit_quants_passive = jnp.append(jax.random.permutation(key_policy, jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4]), independent=True),remainedQuant//4)
                limit_quants = jnp.where(limit_quants_agressive.sum() <= remainedQuant,limit_quants_agressive,limit_quants_passive)
                market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
                quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
                # ---------- quants ----------
                return jnp.array(quants)
            def twapV3(state, env_params):
                # ---------- ifMarketOrder ----------
                remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
                marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
                ifMarketOrder = (remainingTime <= marketOrderTime)
                # print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
                # ---------- ifMarketOrder ----------
                # ---------- quants ----------
                remainedQuant = state.task_to_execute - state.quant_executed
                remainedStep = state.max_steps_in_episode - state.step_counter
                stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
                limit_quants = jax.random.permutation(key_policy, jnp.array([stepQuant//2,stepQuant-stepQuant//2,stepQuant//2,stepQuant-stepQuant//2]), independent=True)
                market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
                quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
                # ---------- quants ----------
                return jnp.array(quants) 
               
            twap_action = twapV3(state, env_params)
            print(f"Sampled {i}th actions are: ",twap_action)
            start=time.time()
            obs,state,reward,done,info=env.step(key_step, state,twap_action, env_params)
            print(f"Time for {i} step: \n",time.time()-start)
            print("excuted ",info["quant_executed"])
            excuted_list.append(info["quant_executed"])
            if done:
                break
        return info['window_index'], info['average_price'], excuted_list
    def get_random_average_price(rngInitNum):
        # ---------- init probabilities ----------
        import numpy as np
        p_0_1 = 0.9 # Define the probabilities for the numbers 0 and 1
        p_2_10 = 0.1 # Define the remaining probability for the numbers 2 through 10
        numbers_2_10 = np.arange(2, 200) # Define the numbers 2 through 10
        pareto_distribution = (1 / numbers_2_10)**4 # Generate a Pareto distribution for the numbers 2 through 200
        pareto_distribution /= pareto_distribution.sum()
        pareto_distribution *= p_2_10
        probabilities = np.array([p_0_1 / 2, p_0_1 / 2] + list(pareto_distribution)) # Combine the probabilities for all numbers from 0 to 10
        assert np.isclose(probabilities.sum(), 1.0) # Verify that the probabilities sum to 1
        # ---------- init probabilities ----------
        rng = jax.random.PRNGKey(rngInitNum)
        rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        print("Time for reset: \n",time.time()-start)
        excuted_list = []
        for i in range(1,10000):
            print("---"*20)
            print("window_index ",state.window_index)
            key_policy, _ = jax.random.split(key_policy,2)
            def randomV1(state, env_params):
                quants = np.random.choice(np.arange(0, 200), size=4, p=probabilities) # Generate random data from the custom distribution
                return jnp.array(quants)                
            random_action = randomV1(state, env_params)
            print(f"Sampled {i}th actions are: ",random_action)
            start=time.time()
            obs,state,reward,done,info=env.step(key_step, state,random_action, env_params)
            print(f"Time for {i} step: \n",time.time()-start)
            print("excuted ",info["quant_executed"])
            excuted_list.append(info["quant_executed"])
            if done:
                break
        return info['window_index'], info['average_price'], excuted_list
    def get_hush_average_price(rngInitNum):
        rng = jax.random.PRNGKey(rngInitNum)
        rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        print("Time for reset: \n",time.time()-start)
        excuted_list = []
        for i in range(1,10000):
            print("---"*20)
            print("window_index ",state.window_index)
            key_policy, _ = jax.random.split(key_policy,2)
            def rushV1(state, env_params):
                quants = np.random.choice(np.arange(0, 200), size=4) # Generate random data from the custom distribution
                return jnp.array(quants)                
            random_action = rushV1(state, env_params)
            print(f"Sampled {i}th actions are: ",random_action)
            start=time.time()
            obs,state,reward,done,info=env.step(key_step, state,random_action, env_params)
            print(f"Time for {i} step: \n",time.time()-start)
            print("excuted ",info["quant_executed"])
            excuted_list.append(info["quant_executed"])
            if done:
                break
        return info['window_index'], info['average_price'], excuted_list
    # def get_best_price_average_price(rngInitNum):
    #     rng = jax.random.PRNGKey(rngInitNum)
    #     rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
    #     start=time.time()
    #     obs,state=env.reset(key_reset,env_params)
    #     print("Time for reset: \n",time.time()-start)
    #     excuted_list = []
    #     for i in range(1,10000):
    #         print("---"*20)
    #         print("window_index ",state.window_index)
    #         key_policy, _ = jax.random.split(key_policy,2)
    #         def bestPrice(state, env_params):
    #             quants = jnp.array([min(remaining Q, size at best price), 0, 0, 0])
    #             return jnp.array(quants)                
    #         random_action = bestPrice(state, env_params)
    #         print(f"Sampled {i}th actions are: ",random_action)
    #         start=time.time()
    #         obs,state,reward,done,info=env.step(key_step, state,random_action, env_params)
    #         print(f"Time for {i} step: \n",time.time()-start)
    #         print("excuted ",info["quant_executed"])
    #         excuted_list.append(info["quant_executed"])
    #         if done:
    #             break
    #     return info['window_index'], info['average_price'], excuted_list
    def get_advantage(rngInitNum):
        window_index2,twap,executed_list2=get_twap_average_price(rngInitNum)
        window_index1,ppo,executed_list1=get_ppo_average_price(rngInitNum)
        window_index3,random,executed_list3=get_random_average_price(rngInitNum)
        window_index4,rush,executed_list4=get_hush_average_price(rngInitNum)
        assert window_index1 == window_index2
        assert window_index1 == window_index3
        assert window_index1 == window_index4
        return window_index1, (ppo-twap)/twap*10000, (ppo-random)/random*10000, (ppo-rush)/rush*10000, ppo, twap, random, rush, executed_list1, executed_list2, executed_list3, executed_list4
    # result_list = []
    for rngInitNum in range(100,1000):
        print(f"++++ rngInitNum {rngInitNum}")
        result_tuple = get_advantage(rngInitNum) 
        # result_list.append(result_tuple[0]) # window index
        print(f"window_index {result_tuple[0]:<4} , advantageTWAP {result_tuple[1]:^20} , advantageRANDOM {result_tuple[2]:^20} , advantageRUSH {result_tuple[3]:^20} , ppoAP {result_tuple[4]:<20} , twapAP {result_tuple[5]:<20} , randomAP {result_tuple[6]:<20} , rushAP {result_tuple[7]:<20} , ppoExecuted { [int(x) for x in result_tuple[8]]} , twapExecuted { [int(x) for x in result_tuple[9]]} , randomExecuted { [int(x) for x in result_tuple[10]]} , rushExecuted { [int(x) for x in result_tuple[11]]}",\
            file=open('comparison'+ paramsFile.split('_')[-3] +"_OneMonth_"+timestamp+'.txt','a'))
        
        
        
        





from functools import partial, partialmethod
from typing import OrderedDict
from jax import numpy as jnp
import jax
import gymnax_exchange.jaxob.JaxOrderBookArrays as job
import time
import timeit


import sys
sys.path.append('/Users/sasrey/AlphaTrade')
import gymnax_exchange


from gym_exchange.data_orderbook_adapter.raw_encoder import RawDecoder, RawEncoder
from gym_exchange.data_orderbook_adapter.decoder import Decoder
from gym_exchange.data_orderbook_adapter.encoder import Encoder
from gym_exchange.data_orderbook_adapter.data_pipeline import DataPipeline




class OrderBook():
    def __init__(self,nOrders=10,nTrades=10) -> None:
        self.nOrders=nOrders
        self.nTrades=nTrades
        self.bids=(jnp.ones((nOrders,6))*-1).astype("int32")
        self.asks=(jnp.ones((nOrders,6))*-1).astype("int32")
        self.trades=(jnp.ones((nTrades,6))*-1).astype("int32")
        

    


    #For loop with jitted functions inside the for loop.
    def process_order_array(self,order_array):
        bidside=self.bids
        askside=self.asks
        trades=self.trades
        ordersides=(askside.astype(jnp.int32),bidside.astype(jnp.int32),trades.astype(jnp.int32))
        for msg in order_array:
            ordersides,_=job.cond_type_side(ordersides,msg)
        self.asks=ordersides[0]
        self.bids=ordersides[1]
        self.trades=ordersides[2]

    def process_order_array_scan(self,order_array):
        bidside=self.bids
        askside=self.asks
        trades=self.trades
        ordersides=job.scan_through_entire_array(order_array,(askside,bidside,trades))
        self.asks=ordersides[0]
        self.bids=ordersides[1]
        self.trades=ordersides[2]
        return ordersides[0],ordersides[1],ordersides[2]


    def vprocess_mult_order_arrays(self,single_array,Nparallel):
        msg_arrays=jnp.stack([single_array]*Nparallel,axis=2)
        bidsides=jnp.stack([self.bids]*Nparallel,axis=0).astype("int32")
        asksides=jnp.stack([self.asks]*Nparallel,axis=0).astype("int32")
        ordersides=(asksides,bidsides)
        print("Msg Batch Shape:",msg_arrays.shape)
        print("Bidside Batch Shape:",bidsides.shape)
        print("Askside Batch Shape:",asksides.shape)
        for msg in msg_arrays:
            ordersides,trades=job.vcond_type_side(ordersides,msg)
        return ordersides,trades

    def vprocess_mult_order_arrays_scan(self,single_array,Nparallel):
        msg_arrays=jnp.stack([single_array]*Nparallel,axis=2)
        bidsides=jnp.stack([self.bids]*Nparallel,axis=0).astype("int32")
        asksides=jnp.stack([self.asks]*Nparallel,axis=0).astype("int32")
        ordersides=(asksides,bidsides,jnp.ones_like(bidsides)*-1)
        print("Msg Batch Shape:",msg_arrays.shape)
        print("Bidside Batch Shape:",bidsides.shape)
        print("Askside Batch Shape:",asksides.shape)
        ordersides=job.vscan_through_entire_array(msg_arrays,ordersides)
        return ordersides


    def get_L2_state(self,N):
        bid_prices=jnp.sort(self.bids[:,0])
        ask_prices=jnp.sort(self.asks[:,0])
        topNbid=bid_prices[-N:][::-1]
        index=int(jnp.where(ask_prices[::-1]==-1,fill_value=0,size=1)[0])
        topNask=jnp.concatenate((ask_prices[-index:],jnp.zeros(N).astype("int32")))
        topNask=topNask[0:N]
        bids=jnp.stack((job.get_totquant_at_prices(self.bids,topNbid),topNbid))
        asks=jnp.stack((job.get_totquant_at_prices(self.asks,topNask),topNask))
        return bids.T,asks.T
    

    #Flatten and Unflatten functions so that methods can be appropriately jitted. 
    def _tree_flatten(self):
        children = (self.bids,self.asks)  # arrays / dynamic values
        aux_data = {'nOrders': self.nOrders,'nTrades':self.nTrades}  # static values
        return (children, aux_data)

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(**aux_data)


jax.tree_util.register_pytree_node(OrderBook,
                                    OrderBook._tree_flatten,
                                    OrderBook._tree_unflatten)

#Turns flow lists to have side be -1/1 rather than ask/bid
def to_order_flow_lists(flow_lists): 
    '''change side format from bid/ask to 1/-1
    side = -1 if item.side == 'ask' else 1'''
    for flow_list in flow_lists:
        for item in flow_list:
            side = -1 if item.side == 'ask' else 1
            item.side = side
    return flow_lists





if __name__ == "__main__":
     ##Configuration variables:
    ordersPerSide=100
    instancesInParallel=1000
    runfirstcall=True
    run2ndcall=True
    runmult_for=False
    runmult_scan=True
    printTimes=True
    printResults=False
    time_individual_msgs=True
    develop=True
    load_data=True
    if load_data:
        #Loading data from the LOBSTER dataset using Kang's data encoder - where all the pre-processing is done. 
        decoder = RawDecoder(**DataPipeline()())
        encoder = RawEncoder(decoder)
        flow_lists=encoder()
        #Function by me to change the format of the flow lists slightly.
        flow_lists = to_order_flow_lists(flow_lists)
        #Single message for debugging purposes. 
        single_message=flow_lists[0].get_head_order_flow().to_message
        jax_list=[]
        message_list=[]

        #Creating two types of lists of messages to give to the orderbook:
        # A) jax_list is a list with messages ready for the jax OB (i.e. numerical)
        for flow_list in flow_lists[0:9]:
            for flow in flow_list:
                jax_list.append(flow.to_list)
                message_list.append(flow.to_message)
        message_array=jnp.array(jax_list)
        #print("Messages processed: \n",message_array)
        #print("1st message: " ,message_array[0])


   


    print("Processing...")
    #Process a single set of messages (message_array) through the orderbook. This is largely to make sure everything is jitted and d
    if runfirstcall:
        ob=OrderBook(nOrders=ordersPerSide,nTrades=10)
        ob.process_order_array_scan(message_array)
        if printResults:
            print(message_array)
            print(ob.trades)
        #ob_branch=OrderBook(nOrders=ordersPerSide)
        #ob_branch.process_order_array_branch(message_array)

    start=time.time()
    #Process a second set of messages in a seperate orderbook object. 
    if run2ndcall:
        ob2=OrderBook(nOrders=ordersPerSide)
        #Include timing
        start=time.time()
        ob2.process_order_array(message_array)   
        #Print the output state of both sides of the orderbook.
        if printResults:
            print("Asks: \n",ob2.asks)
            print("Bids: \n",ob2.bids)
    end_single=time.time()-start


    #Process a set of messages N times in parallel, with a for loop going through messages sequentially but processing all N orderbooks in parallel.
    if runmult_for:
        ob4=OrderBook(nOrders=ordersPerSide)
        start=time.time()
        val=ob4.vprocess_mult_order_arrays(message_array,instancesInParallel)
        if printResults:
            print(jax.tree_util.tree_structure(val))
    end_for=time.time()-start


    if runmult_scan:
        ob3=OrderBook(nOrders=ordersPerSide)
        start=time.time()
        val=ob3.vprocess_mult_order_arrays_scan(message_array,instancesInParallel)
        if printResults:
            print(val)
            print(jax.tree_util.tree_structure(val))
    end_scan=time.time()-start
    
    if printTimes:
        print("Time required for N=",1," single instance with cond:", end_single)
        print("Time for for loop with N=",instancesInParallel," instances with cond in a for loop (vmap inside)", end_for)
        print("Time for scan with N=",instancesInParallel," instances with cond in a scan (vmap outside)", end_scan)

    if time_individual_msgs:
        #test_msg=jnp.array([1,1,10,10,0,0,5,5])
        lim_msg={
        'side':-1,
        'type':1,
        'price':10000,
        'quantity':10,
        'orderid':0,
        'traderid':1,
        'time':34000,
        'time_ns':78}

        cncl_msg={
        'side':1,
        'type':3,
        'price':10000,
        'quantity':10,
        'orderid':0,
        'traderid':1,
        'time':34000,
        'time_ns':90}


        match_msg={
        'side':1,
        'type':1,
        'price':11000,
        'quantity':100,
        'orderid':0,
        'traderid':1,
        'time':34000,
        'time_ns':89897}


        b=(jnp.ones((100,6))*-1).astype(jnp.int32)
        a=(jnp.ones((100,6))*-1).astype(jnp.int32)
        t=(jnp.ones((1000,6))*-1).astype(jnp.int32)

        returnval=job.ask_lim(lim_msg,a,b,t)



        #job.ask_lim(msg,ask,bid,trades)
        n_runs=10000
        print("Limit time:",timeit.timeit('val=job.ask_lim(lim_msg,a,b,t); jax.block_until_ready(val)',number=n_runs,globals=globals())/n_runs)

        a,b,t=returnval
        returnval=job.ask_lim(lim_msg,a,b,t)
        a,b,t=returnval
        returnval=job.ask_lim(lim_msg,a,b,t)
        a,b,t=returnval
        returnval=job.ask_lim(lim_msg,a,b,t)
        a,b,t=returnval
        returnval=job.ask_cancel(cncl_msg,a,b,t)   
        print("Cancel time:",timeit.timeit('val=job.ask_cancel(cncl_msg,a,b,t); jax.block_until_ready(val)',number=n_runs,globals=globals())/n_runs)

        a,b,t=returnval
        returnval=job.bid_lim(match_msg,a,b,t)
        print("Matching time:",timeit.timeit('val=job.bid_lim(match_msg,a,b,t); jax.block_until_ready(val)',number=n_runs,globals=globals())/n_runs)
        #print(returnval)

    if develop:
        print('developing')


# ============== testing scripts ===============
import jax
import jax.numpy as jnp
import sys
from re import L
import time 
sys.path.append('/Users/sasrey/AlphaTrade')
sys.path.append('/homes/80/kang/AlphaTrade')
import chex
import faulthandler; faulthandler.enable()
chex.assert_gpu_available(backend=None)
from jax import config # Code snippet to disable all jitting.
config.update("jax_disable_jit", False)
# config.update("jax_disable_jit", True)
from gymnax_exchange.jaxen.exec_env import *
import json
# ============== testing scripts ===============

if __name__ == "__main__":
    try:
        ATFolder = sys.argv[1]
        print("AlphaTrade folder:",ATFolder)
    except:
        # ATFolder = '/home/duser/AlphaTrade'
        # ATFolder = '/homes/80/kang/AlphaTrade'
        # ATFolder = '/homes/80/kang/AlphaTrade/testing'
        # ATFolder = '/homes/80/kang/AlphaTrade/testing_small'
        ATFolder = "/homes/80/kang/AlphaTrade/training_oneDay"
    env=ExecutionEnv(ATFolder,"sell")
    env_params=env.default_params
    print(env_params.message_data.shape, env_params.book_data.shape)
    assert env.task_size == 500    
    
    import time;timestamp = str(int(time.time()))
    rngInitNum = 0;rng = jax.random.PRNGKey(rngInitNum)
    
    
    
    enable_vmap=True
                
    if enable_vmap:
        vmap_reset = jax.jit(jax.vmap(env.reset, in_axes=(0, None)))
        vmap_step = jax.jit(jax.vmap(env.step, in_axes=(0, 0, 0, None)))
        vmap_act_sample=jax.jit(jax.vmap(env.action_space().sample, in_axes=(0)))
        def twapV3(action_key, state, env_params):
            # ---------- ifMarketOrder ----------
            remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
            marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
            ifMarketOrder = (remainingTime <= marketOrderTime)
            # print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
            # ---------- ifMarketOrder ----------
            # ---------- quants ----------
            remainedQuant = state.task_to_execute - state.quant_executed
            remainedStep = state.max_steps_in_episode - state.step_counter
            stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
            limit_quants = jax.random.permutation(action_key, jnp.array([stepQuant//2,stepQuant-stepQuant//2,stepQuant//2,stepQuant-stepQuant//2]), independent=True)
            market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
            quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
            # ---------- quants ----------
            return jnp.array(quants)
        vmap_act_twapV3 = jax.jit(jax.vmap(twapV3,in_axes=(0, 0, None)))

        globalSteps = int(1e7)
        num_envs = 1000
        # globalSteps = 4*500
        # num_envs = 4
        n_steps=globalSteps//num_envs
        print(n_steps)
        vmap_keys = jax.random.split(rng, num_envs)

        start=time.time()
        obs, state = vmap_reset(vmap_keys, env_params)
        print("Time for vmap reset with,",num_envs, " environments : \n",time.time()-start)

        start=time.time()
        n_obs, n_state, reward, done, _ = vmap_step(vmap_keys, state, vmap_act_sample(vmap_keys), env_params)
        print("Time for vmap step with,",num_envs, " environments : \n",time.time()-start)

        # def step_wrap_vmap(runner_state, unused):
        #     jax.debug.print('@step')
        def step_wrap_vmap(runner_state, i):
            jax.debug.print('step {}',i)
            env_state, obsv, done, rng, cum_reward = runner_state
            rng, _rng = jax.random.split(rng)
            vmap_keys = jax.random.split(_rng, num_envs)

            test_actions = vmap_act_twapV3(vmap_keys,env_state,env_params)
            # test_actions = vmap_act_sample(vmap_keys).astype(jnp.float32)
            obsv, env_state, reward, done, info = vmap_step(vmap_keys, env_state, test_actions, env_params)
            cum_reward += reward
            reset_reward = jnp.where(done, 0.0, cum_reward)  # Reset cum_reward to 0 if done is True
            runner_state = (env_state, obsv, done, rng, reset_reward)
            # jax.debug.print("reward {}", reward)
            return runner_state, (cum_reward, done)
        
        initial_cum_reward = jnp.zeros((num_envs,))
        initial_dones = jnp.zeros((num_envs,), dtype=jnp.bool_)
        r_state = (n_state, n_obs, initial_dones, rng, initial_cum_reward)

        # def scan_func_vmap(r_state, n_steps):
        #     r_state, (cum_rewards, dones) = jax.lax.scan(step_wrap_vmap, r_state, None, n_steps)
        #     return r_state, (cum_rewards, dones)
        
        def scan_func_vmap(r_state, n_steps):
            # Generating a sequence of iteration numbers
            iterations = jnp.arange(n_steps)
            r_state, (cum_rewards, dones) = jax.lax.scan(step_wrap_vmap, r_state, iterations)
            return r_state, (cum_rewards, dones)



        start = time.time()
        r_state, (cum_rewards, dones) = jax.jit(scan_func_vmap, static_argnums=(1,))(r_state, n_steps)
        print("Time for vmap step with,", num_envs, " environments and", n_steps, " steps: \n", time.time() - start)

        adjusted_rewards = cum_rewards * dones.astype(jnp.float32)
        adjusted_rewards
        adjusted_rewards.shape
        
        
        
        # Assuming adjusted_rewards is a jnp.array
        non_zero_mask = adjusted_rewards != 0
        # Sum of non-zero elements for each row
        non_zero_sums = jnp.sum(adjusted_rewards * non_zero_mask, axis=1)
        # Count of non-zero elements for each row
        non_zero_counts = jnp.sum(non_zero_mask, axis=1)
        # Avoid division by zero by replacing zero counts with 1
        non_zero_counts = jnp.where(non_zero_counts == 0, 1, non_zero_counts)
        # Mean of non-zero elements for each row
        non_zero_means = non_zero_sums / non_zero_counts
        non_zero_means
        non_zero_means.shape
        
        
        # Assuming `sample_array` is your input jnp.array
        non_zero_indices = jnp.nonzero(non_zero_means)[0]
        non_zero_values = non_zero_means[non_zero_indices]
        index_value_pairs = list(zip(non_zero_indices, non_zero_values))
        import pandas as pd
        df = pd.DataFrame(index_value_pairs)
        df.to_csv("twap.csv")

        
        
        
        
        # for i in range(adjusted_rewards.shape[0]):
        #     print(adjusted_rewards[i,:])
                
    # enable_vmap=True
    # if enable_vmap:
    #     # with jax.profiler.trace("/homes/80/kang/AlphaTrade/wandb/jax-trace"):
    #     vmap_reset = jax.vmap(env.reset, in_axes=(0, None))
        
    #     vmap_step = jax.vmap(env.step, in_axes=(0, 0, 0, None))
    #     vmap_act_sample=jax.vmap(env.action_space().sample, in_axes=(0))
    #     # print(test_actions)

    #     num_envs = 10
    #     rng, resetKey, actionKey, stepKey = jax.random.split(rng, 4)


    #     start=time.time()
    #     resetKeys =jax.random.split(resetKey, num_envs)
    #     actionKeys =jax.random.split(actionKey, num_envs)
    #     stepKeys =jax.random.split(stepKey, num_envs)
    #     obs, state = vmap_reset(resetKeys, env_params)
    #     print("Time for vmap reset with,",num_envs, " environments : \n",time.time()-start)
    #     total_steps = 1000*10
    #     for i in range(total_steps//num_envs):
    #         start=time.time()
    #         actionKeys = jax.random.split(actionKeys[0], num_envs)
    #         stepKeys = jax.random.split(stepKeys[0], num_envs)
    #         test_actions=vmap_act_sample(actionKeys)
    #         n_obs, n_state, n_reward, n_done, n_info = vmap_step(stepKeys, state, test_actions, env_params)
    #         print("Time for vmap step with,",num_envs, " environments : \n",time.time()-start)
    #         print(f"\n-------- {i} --------")
    #         print(test_actions)
    #         print(n_done)
    #         print("stepKeys",stepKeys)
    #         print("current_step",n_info["current_step"])
    #         print("window_index",n_info["window_index"])
    #         if any(n_done):
    #             break
    #         # if done:
    #         #     vmap_keys = jax.random.split(vmap_keys, num_envs)
    #         #     start=time.time()
    #         #     obs, state = vmap_reset(vmap_keys, env_params)
    #         #     print("Time for vmap reset with,",num_envs, " environments : \n",time.time()-start)                
                
                
    # rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)
    # start=time.time()
    # obs,state=env.reset(key_reset,env_params)
    # print("Time for reset: \n",time.time()-start)
    # print(env_params.message_data.shape, env_params.book_data.shape)
    # erlist = []
    # excuted_list = []
    # er = 0
    # for i in range(1,int(1e7)):
    #     # ==================== ACTION ====================
    #     # ---------- acion from given strategy  ----------
    #     print("---"*20)
    #     print("window_index ",state.window_index)
    #     key_policy, _ = jax.random.split(key_policy,2)
    #     def twap(state, env_params):
    #         # ---------- ifMarketOrder ----------
    #         remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
    #         marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
    #         ifMarketOrder = (remainingTime <= marketOrderTime)
    #         print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
    #         # ---------- ifMarketOrder ----------
    #         # ---------- quants ----------
    #         remainedQuant = state.task_to_execute - state.quant_executed
    #         remainedStep = state.max_steps_in_episode - state.step_counter
    #         stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
    #         limit_quants_agressive = jnp.append(jax.random.permutation(key_policy, jnp.array([stepQuant - 3*stepQuant//4,stepQuant//4, stepQuant//4]), independent=True),max(stepQuant - 3*stepQuant//4,stepQuant//4))
    #         limit_quants_passive = jnp.append(jax.random.permutation(key_policy, jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4]), independent=True),remainedQuant//4)
    #         limit_quants = jnp.where(limit_quants_agressive.sum() <= remainedQuant,limit_quants_agressive,limit_quants_passive)
    #         market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
    #         quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
    #         # ---------- quants ----------
    #         return jnp.array(quants)
    #     def twapV3(state, env_params):
    #         # ---------- ifMarketOrder ----------
    #         remainingTime = env_params.episode_time - jnp.array((state.time-state.init_time)[0], dtype=jnp.int32)
    #         marketOrderTime = jnp.array(60, dtype=jnp.int32) # in seconds, means the last minute was left for market order
    #         ifMarketOrder = (remainingTime <= marketOrderTime)
    #         # print(f"{i} remainingTime{remainingTime} marketOrderTime{marketOrderTime}")
    #         # ---------- ifMarketOrder ----------
    #         # ---------- quants ----------
    #         remainedQuant = state.task_to_execute - state.quant_executed
    #         remainedStep = state.max_steps_in_episode - state.step_counter
    #         stepQuant = jnp.ceil(remainedQuant/remainedStep).astype(jnp.int32) # for limit orders
    #         limit_quants = jax.random.permutation(key_policy, jnp.array([stepQuant//2,stepQuant-stepQuant//2,stepQuant//2,stepQuant-stepQuant//2]), independent=True)
    #         market_quants = jnp.array([remainedQuant - 3*remainedQuant//4,remainedQuant//4, remainedQuant//4, remainedQuant//4])
    #         quants = jnp.where(ifMarketOrder,market_quants,limit_quants)
    #         # ---------- quants ----------
    #         return jnp.array(quants) 
            
    #     twap_action = twapV3(state, env_params)
    #     print(f"Sampled {i}th actions are: ",twap_action)
    #     start=time.time()
    #     obs,state,reward,done,info=env.step(key_step, state,twap_action, env_params)
    #     er+=reward
    #     print(f"Time for {i} step: \n",time.time()-start)
    #     print("excuted ",info["quant_executed"])
    #     excuted_list.append(info["quant_executed"])
    #     if done:
    #         erlist.append((i, er))
    #         print(f"global step {i:<10} , episodic return {er:^20} , ",\
    #             file=open('twap_'+ timestamp +"_OneDay_train_"+'.txt','a'))
    #         key_reset, _ = jax.random.split(key_reset)
    #         obs,state=env.reset(key_reset,env_params)
    #         er =0
    
    


# ===================================== #
# ******* Config your own func ******** #
# ===================================== #

from gymnax_exchange.jaxob.JaxOrderBookArrays import *
@partial(jax.jit)
def get_best_bid(asks, bids):
    L2_state = get_L2_state(1, asks, bids)
    return L2_state[0]

@partial(jax.jit)
def get_best_bid(asks, bids):
    L2_state = get_L2_state(1, asks, bids)
    return L2_state[1]



import jax
import jax.numpy as jnp
import gymnax
import sys
sys.path.append('/Users/sasrey/AlphaTrade')
sys.path.append('/homes/80/kang/AlphaTrade')
sys.path.append('../purejaxrl')

from gymnax_exchange.jaxen.exec_env import ExecutionEnv
from gymnax_exchange.jaxob import JaxOrderBookArrays as job
import chex
import time
from purejaxrl.wrappers import FlattenObservationWrapper, LogWrapper,ClipAction, VecEnv,NormalizeVecObservation,NormalizeVecReward


import faulthandler

faulthandler.enable()
print("Num Jax Devices:",jax.device_count(),"Device List:",jax.devices())

from jax import config 
config.update("jax_check_tracer_leaks",False)


chex.assert_gpu_available(backend=None)

#Code snippet to disable all jitting.
#from jax import config
#config.update("jax_disable_jit", True)

print("#"*20+"Output Log File"+"#"*20,file=open('output.txt','w'))

if __name__ == "__main__":
    try:
        ATFolder = sys.argv[1] 
    except:
        ATFolder = '/homes/80/kang/AlphaTrade'
    print("AlphaTrade folder:",ATFolder)

    enable_vmap=True 
    enable_simple=True

    rng = jax.random.PRNGKey(0)
    rng, key_reset, key_policy, key_step = jax.random.split(rng, 4)

    env=ExecutionEnv(ATFolder,'buy',True,'absolute')
    env_params=env.default_params
    env = LogWrapper(env)
    print('Shape of message data and book data',env_params.message_data.shape, env_params.book_data.shape)
    
    #print(env_params.message_data[0,89:95,:,0])
    
    if enable_simple:

        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        print("State after reset: \n",state,file=open('output.txt','a'))
        print("Observation after reset: \n",obs,file=open('output.txt','a'))
        print("Time for reset: \n",time.time()-start)
        print(env_params.message_data.shape, env_params.book_data.shape)

        
        start=time.time()
        obs,state=env.reset(key_reset,env_params)
        print("Time for 2nd reset: \n",time.time()-start)
        print(env_params.message_data.shape, env_params.book_data.shape)

        #print(job.get_data_messages(env_params.message_data,state.window_index,state.step_counter+1))

        #print(env.action_space().sample(key_policy))
        #print(env.state_space(env_params).sample(key_policy))


        """test_action={"sides":jnp.array([1,1,1]),
                    "quantities":jnp.array([10,10,10]),
                    "prices":jnp.array([2154900,2154000,2153900]),
                    }"""
        
        test_action=env.action_space().sample(key_policy)
        print("Sampled actions are: ",test_action)

        start=time.time()
        obs,state,reward,done,info=env.step(key_step, state,test_action, env_params)
        print("Time for one step: \n",time.time()-start)
        print("State after one step: \n",state,done,file=open('output.txt','a'))
        print("Done after one step: ",done,file=open('output.txt','a'))
        print("Observation after one step: \n",obs,file=open('output.txt','a'))
        

        test_action=env.action_space().sample(key_policy)
        print("Sampled actions are: \n",test_action)
        
        start=time.time()
        obs,state,reward,done,info=env.step(key_step, state,test_action, env_params)
        print("Time for 2nd step: \n",time.time()-start)
        with jnp.printoptions(threshold=jnp.inf):
            print("State after 2 steps: \n",state,done,file=open('output.txt','a'))
            print("Done after 2 steps: ",done,file=open('output.txt','a'))
            print("Observation after 2 steps: \n",obs,file=open('output.txt','a'))
        
        
        start=time.time()
        obs,state,reward,done,info=env.step(key_step, state,test_action, env_params)
        print("Time for 3rd step: \n",time.time()-start)
        with jnp.printoptions(threshold=jnp.inf):
            print("State after 3 steps: \n",state,done,file=open('output.txt','a'))
            print("Done after 3 steps: ",done,file=open('output.txt','a'))
            print("Observation after 3 steps: \n",obs,file=open('output.txt','a'))
        
        #comment



        ####### Testing the vmap abilities ########
    

    if enable_vmap:
        vmap_reset = jax.jit(jax.vmap(env.reset, in_axes=(0, None)))
        vmap_step = jax.jit(jax.vmap(env.step, in_axes=(0, 0, 0, None)))
        vmap_act_sample=jax.jit(jax.vmap(env.action_space().sample, in_axes=(0)))

        num_envs = 10000
        vmap_keys = jax.random.split(rng, num_envs)

        test_actions=vmap_act_sample(vmap_keys)
        print(test_actions)

        start=time.time()
        obs, state = vmap_reset(vmap_keys, env_params)
        print("Time for vmap reset with,",num_envs, " environments : \n",time.time()-start)

        start=time.time()
        n_obs, n_state, reward, done, _ = vmap_step(vmap_keys, state, test_actions, env_params)
        print("Time for vmap step with,",num_envs, " environments : \n",time.time()-start)

        def step_wrap_vmap(runner_state,unused):
            jax.debug.print('step')
            env_state, obsv, done, rng=runner_state
            rng,_rng=jax.random.split(rng)
            vmap_keys = jax.random.split(_rng, num_envs)
            test_actions=vmap_act_sample(vmap_keys).astype(jnp.float32)
            obsv,env_state,reward,done,info=vmap_step(vmap_keys, env_state, test_actions, env_params)
            runner_state = (env_state, obsv, done, rng)
            return runner_state,None
    
        r_state=(n_state, n_obs, done, rng)


        n_steps=500


        def scan_func_vmap(r_state,n_steps):
            r_state,_=jax.lax.scan(step_wrap_vmap,r_state,None,n_steps)
            return r_state

        start=time.time()
        r_state=jax.jit(scan_func_vmap,static_argnums=(1,))(r_state,n_steps)
        print("Time for vmap step with,",num_envs, " environments and",n_steps," steps: \n",time.time()-start)